import dataclasses
from collections.abc import Callable, Mapping
from typing import Any, Literal

import torch
import torch.linalg

from fme.core.device import get_device
from fme.core.ensemble import get_crps, get_energy_score
from fme.core.gridded_ops import GriddedOperations
from fme.core.normalizer import StandardNormalizer
from fme.core.packer import Packer
from fme.core.typing_ import TensorMapping


class NaNLoss(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, input, target):
        return torch.tensor(torch.nan)


class WeightedMappingLoss:
    def __init__(
        self,
        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
        weights: dict[str, float],
        out_names: list[str],
        normalizer: StandardNormalizer,
        channel_dim: int = -3,
    ):
        """
        Args:
            loss: The loss function to apply.
            weights: A dictionary of variable names with individual
                weights to apply to their normalized losses
            out_names: The names of the output variables.
            normalizer: The normalizer to use.
            channel_dim: The channel dimension of the input tensors.
        """
        self._weight_tensor = _construct_weight_tensor(
            weights, out_names, channel_dim=channel_dim
        )
        self.loss = VariableWeightingLoss(
            weights=self._weight_tensor,
            loss=loss,
        )
        if self._weight_tensor.flatten().shape[0] != len(out_names):
            raise RuntimeError(
                "The number of weights must match the number of output names, "
                "behavior of _construct_weight_tensor has changed."
            )
        self.packer = Packer(out_names)
        self.channel_dim = channel_dim
        self.normalizer = normalizer

    def __call__(
        self,
        predict_dict: TensorMapping,
        target_dict: TensorMapping,
    ):
        predict_tensors = self.packer.pack(
            self.normalizer.normalize(predict_dict), axis=self.channel_dim
        )
        target_tensors = self.packer.pack(
            self.normalizer.normalize(target_dict), axis=self.channel_dim
        )
        nan_mask = target_tensors.isnan()
        if nan_mask.any():
            predict_tensors = torch.where(nan_mask, 0.0, predict_tensors)
            target_tensors = torch.where(nan_mask, 0.0, target_tensors)

        return self.loss(predict_tensors, target_tensors)

    def get_normalizer_state(self) -> dict[str, float]:
        return self.normalizer.get_state()

    @property
    def effective_loss_scaling(self) -> dict[str, float]:
        custom_weights = dict(zip(self.packer.names, self._weight_tensor.flatten()))
        loss_normalizer_stds = self.normalizer.stds
        return {
            k: loss_normalizer_stds[k] / custom_weights[k] for k in self.packer.names
        }


def _construct_weight_tensor(
    weights: dict[str, float],
    out_names: list[str],
    n_dim: int = 4,
    channel_dim: int = -3,
) -> torch.Tensor:
    """Creates a packed weight tensor with the appropriate dimensions for
    broadcasting with generated or target output tensors. When used in
    the n_forward_steps loop in the stepper's run_on_batch, the channel dim is
    -3 and the n_dim is 4 (sample, channel, lat, lon).

    Args:
        weights: dict of variable names with individual weights to apply
            to their normalized loss
        out_names: list of output variable names
        n_dim: number of dimensions of the output tensor
        channel_dim: the channel dimension of the output tensor
    """
    weights_tensor = torch.tensor([weights.get(key, 1.0) for key in out_names])
    # positive index of the channel dimension
    _channel_dim = n_dim + channel_dim if channel_dim < 0 else channel_dim
    reshape_dim = (
        len(weights_tensor) if i == _channel_dim else 1 for i in range(n_dim)
    )
    return weights_tensor.reshape(*reshape_dim).to(get_device(), dtype=torch.float)


class LpLoss(torch.nn.Module):
    def __init__(self, p=2):
        """
        Args:
            p: Lp-norm type. For example, p=1 for L1-norm, p=2 for L2-norm.
        """
        super().__init__()

        if p <= 0:
            raise ValueError("Lp-norm type should be positive")

        self.p = p

    def rel(self, x, y):
        num_examples = x.size()[0]

        diff_norms = torch.linalg.norm(
            x.reshape(num_examples, -1) - y.reshape(num_examples, -1), ord=self.p, dim=1
        )
        y_norms = torch.linalg.norm(y.reshape(num_examples, -1), ord=self.p, dim=1)

        return torch.mean(diff_norms / y_norms)

    def __call__(self, x, y):
        return self.rel(x, y)


class AreaWeightedMSELoss(torch.nn.Module):
    def __init__(self, area_weighted_mean: Callable[[torch.Tensor], torch.Tensor]):
        super().__init__()
        self._area_weighted_mean = area_weighted_mean

    def __call__(self, x, y):
        return torch.mean(self._area_weighted_mean((x - y) ** 2))


class WeightedSum(torch.nn.Module):
    """
    A module which applies multiple loss-function modules (taking two inputs)
    to the same input and returns a tensor equal to the weighted sum of the
    outputs of the modules.
    """

    def __init__(self, modules: list[torch.nn.Module], weights: list[float]):
        """
        Args:
            modules: A list of modules, each of which takes two tensors and
                returns a scalar tensor.
            weights: A list of weights to apply to the outputs of the modules.
        """
        super().__init__()
        if len(modules) != len(weights):
            raise ValueError("modules and weights must have the same length")
        self._wrapped = modules
        self._weights = weights

    def __call__(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return sum(w * module(x, y) for w, module in zip(self._weights, self._wrapped))


class GlobalMeanLoss(torch.nn.Module):
    """
    A module which computes a loss on the global mean of each sample.
    """

    def __init__(
        self,
        area_weighted_mean: Callable[[torch.Tensor], torch.Tensor],
        loss: torch.nn.Module,
    ):
        """
        Args:
            area_weighted_mean: Computes an area-weighted mean, removing the
                horizontal dimensions.
            loss: A loss function which takes two tensors of shape
                (n_samples, n_timesteps, n_channels) and returns a scalar
                tensor.
        """
        super().__init__()
        self.global_mean = GlobalMean(area_weighted_mean)
        self.loss = loss

    def __call__(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        x = self.global_mean(x)
        y = self.global_mean(y)
        return self.loss(x, y)


class GlobalMean(torch.nn.Module):
    def __init__(self, area_weighted_mean: Callable[[torch.Tensor], torch.Tensor]):
        """
        Args:
            area_weighted_mean: Computes an area-weighted mean, removing the
                horizontal dimensions.
        """
        super().__init__()
        self._area_weighted_mean = area_weighted_mean

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Args:
            x: A tensor with spatial dimensions in shape (n_samples, n_timesteps,
             n_channels, n_lat, n_lon).
        """
        return self._area_weighted_mean(x)


class VariableWeightingLoss(torch.nn.Module):
    def __init__(self, weights: torch.Tensor, loss: torch.nn.Module):
        """
        Args:
            weights: A tensor of shape (n_samples, n_channels, n_lat, n_lon)
                containing the weights to apply to each channel.
            loss: A loss function which takes two tensors.
        """
        super().__init__()
        self.loss = loss
        self.weights = weights

    def __call__(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return self.loss(self.weights * x, self.weights * y)


class EnergyScoreLoss(torch.nn.Module):
    """
    Compute the energy score over the complex-valued spectral coefficients.

    The energy score is defined as

    .. math::

        E[||X - y||^{beta}] - 1/2 E[||X - X'||^{beta}]

    where :math:`X` is the ensemble, :math:`y` is the target, and :math:`||.||`
    is the complex modulus. It is a proper scoring rule for beta in (0, 2). Here
    we use beta=1. See Gneiting and Raftery (2007) [1]_ Section 4.3 for more details.

    We use a scaling factor of 2 * sqrt(n_l * n_m) to bring its magnitude in
    line with the real-valued CRPS loss, and to prevent its value depending on domain
    size for Gaussian distributed random data where n_lon = 2 * n_lat.

    .. [1] https://sites.stat.washington.edu/people/raftery/Research/PDF/Gneiting2007jasa.pdf
    """

    def __init__(self, sht: Callable[[torch.Tensor], torch.Tensor]):
        super().__init__()
        self.sht = sht
        self.scaling: torch.Tensor | None = None
        self.mode_weights: torch.Tensor | None = None

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        x_hat = self.sht(x)
        y_hat = self.sht(y)
        if self.scaling is None:
            self.scaling = 2 * (x_hat.shape[-2] * x_hat.shape[-1]) ** 0.5
        if self.mode_weights is None:
            # we need to weight the modes properly,
            # with each m mode contributing twice for m!=0:
            H, W = x_hat.shape[-2:]
            self.mode_weights = 2 * torch.ones(
                (*([1] * (x_hat.ndim - 1)), H, W),
                device=x_hat.device,
            )
            self.mode_weights[..., 0] = 1
        return (
            (get_energy_score(x_hat, y_hat) * self.mode_weights)
            .sum(dim=(-2, -1))
            .mean()
        ) / self.scaling


class CRPSLoss(torch.nn.Module):
    """
    Compute the CRPS loss.

    Supports almost-fair modification to CRPS from
    https://arxiv.org/html/2412.15832v1, which claims to be helpful in
    avoiding numerical issues with fair CRPS.
    """

    def __init__(self, alpha: float):
        super().__init__()
        self.alpha = alpha

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return get_crps(x, y, alpha=self.alpha).mean()


class EnsembleLoss(torch.nn.Module):
    def __init__(
        self,
        crps_weight: float,
        energy_score_weight: float,
        sht: Callable[[torch.Tensor], torch.Tensor],
    ):
        super().__init__()
        if crps_weight < 0 or energy_score_weight < 0:
            raise ValueError(
                "crps_weight and energy_score_weight must be non-negative, "
                f"got {crps_weight} and {energy_score_weight}"
            )
        if crps_weight + energy_score_weight == 0:
            raise ValueError(
                "crps_weight and energy_score_weight must sum to a positive value, "
                f"got {crps_weight} and {energy_score_weight}"
            )
        self.crps_loss = CRPSLoss(alpha=0.95)
        self.energy_score_loss = EnergyScoreLoss(sht=sht)

        self.crps_weight = crps_weight
        self.energy_score_weight = energy_score_weight

    def forward(
        self,
        gen_norm: torch.Tensor,
        target_norm: torch.Tensor,
    ):
        if self.crps_weight > 0:
            crps = self.crps_weight * self.crps_loss(gen_norm, target_norm)
        else:
            crps = torch.tensor(0.0)
        if self.energy_score_weight > 0:
            energy_score_loss = self.energy_score_weight * self.energy_score_loss(
                gen_norm, target_norm
            )
        else:
            energy_score_loss = torch.tensor(0.0)
        return crps + energy_score_loss


@dataclasses.dataclass
class LossConfig:
    """
    A dataclass containing all the information needed to build a loss function,
    including the type of the loss function and the data needed to build it.

    Args:
        type: the type of the loss function
        kwargs: data for a loss function instance of the indicated type
        global_mean_type: the type of the loss function to apply to the global
            mean of each sample, by default no loss is applied
        global_mean_kwargs: data for a loss function instance of the indicated
            type to apply to the global mean of each sample
        global_mean_weight: the weight to apply to the global mean loss
            relative to the main loss
    """

    type: Literal["LpLoss", "L1", "MSE", "AreaWeightedMSE", "NaN", "EnsembleLoss"] = (
        "MSE"
    )
    kwargs: Mapping[str, Any] = dataclasses.field(default_factory=lambda: {})
    global_mean_type: Literal["LpLoss"] | None = None
    global_mean_kwargs: Mapping[str, Any] = dataclasses.field(
        default_factory=lambda: {}
    )
    global_mean_weight: float = 1.0

    def __post_init__(self):
        if self.type not in (
            "LpLoss",
            "L1",
            "MSE",
            "AreaWeightedMSE",
            "NaN",
            "EnsembleLoss",
        ):
            raise NotImplementedError(self.type)
        if self.global_mean_type is not None and self.global_mean_type != "LpLoss":
            raise NotImplementedError(self.global_mean_type)

    def build(
        self,
        reduction: Literal["mean", "none"],
        gridded_operations: GriddedOperations | None,
    ) -> Any:
        """
        Args:
            reduction: The reduction to apply to the loss, either "mean" or "none".
                Only used if the loss function is L1, MSE, or LpLoss.
            gridded_operations: The gridded operations to use in the case that
                the loss function requires use of the horizontal dimensions.
        """
        if self.type == "LpLoss":
            main_loss = LpLoss(**self.kwargs)
        elif self.type == "L1":
            main_loss = torch.nn.L1Loss(reduction=reduction)
        elif self.type == "MSE":
            main_loss = torch.nn.MSELoss(reduction=reduction)
        elif self.type == "AreaWeightedMSE":
            if gridded_operations is None:
                raise ValueError("gridded_operations is required for AreaWeightedMSE")
            main_loss = AreaWeightedMSELoss(gridded_operations.area_weighted_mean)
        elif self.type == "NaN":
            main_loss = NaNLoss()
        elif self.type == "EnsembleLoss":
            if gridded_operations is None:
                raise ValueError("gridded_operations is required for EnsembleLoss")
            kwargs = dict(self.kwargs)
            crps_weight = kwargs.pop("crps_weight", 1.0)
            energy_score_weight = kwargs.pop("energy_score_weight", 0.0)
            main_loss = EnsembleLoss(
                sht=gridded_operations.get_real_sht(),
                crps_weight=crps_weight,
                energy_score_weight=energy_score_weight,
                **kwargs,
            )

        if self.global_mean_type is not None:
            if gridded_operations is None:
                raise ValueError("gridded_operations is required for global mean loss")
            global_mean_loss = GlobalMeanLoss(
                area_weighted_mean=gridded_operations.area_weighted_mean,
                loss=LpLoss(**self.global_mean_kwargs),
            )
            final_loss = WeightedSum(
                modules=[main_loss, global_mean_loss],
                weights=[1.0, self.global_mean_weight],
            )
        else:
            final_loss = main_loss
        return final_loss.to(device=get_device())


@dataclasses.dataclass
class WeightedMappingLossConfig:
    """
    Loss configuration class that has the same fields as LossConfig but also
    has additional weights field. The build method will apply the weights to
    the inputs of the loss function. The loss returned by build will be a
    MappingLoss, which takes Dict[str, tensor] as inputs instead of packed
    tensors.

    Args:
        type: the type of the loss function
        kwargs: data for a loss function instance of the indicated type
        global_mean_type: the type of the loss function to apply to the global
            mean of each sample, by default no loss is applied
        global_mean_kwargs: data for a loss function instance of the indicated
            type to apply to the global mean of each sample
        global_mean_weight: the weight to apply to the global mean loss
            relative to the main loss
        weights: A dictionary of variable names with individual
            weights to apply to their normalized losses
    """

    type: Literal["LpLoss", "MSE", "AreaWeightedMSE", "EnsembleLoss"] = "MSE"
    kwargs: Mapping[str, Any] = dataclasses.field(default_factory=lambda: {})
    global_mean_type: Literal["LpLoss"] | None = None
    global_mean_kwargs: Mapping[str, Any] = dataclasses.field(
        default_factory=lambda: {}
    )
    global_mean_weight: float = 1.0
    weights: dict[str, float] = dataclasses.field(default_factory=lambda: {})

    def __post_init__(self):
        self.loss_config = LossConfig(
            type=self.type,
            kwargs=self.kwargs,
            global_mean_type=self.global_mean_type,
            global_mean_kwargs=self.global_mean_kwargs,
            global_mean_weight=self.global_mean_weight,
        )

    def build(
        self,
        gridded_ops: GriddedOperations | None,
        out_names: list[str],
        normalizer: StandardNormalizer,
        channel_dim: int = -3,
    ) -> WeightedMappingLoss:
        loss = self.loss_config.build(
            reduction="mean",
            gridded_operations=gridded_ops,
        )
        return WeightedMappingLoss(
            loss=loss,
            weights=self.weights,
            out_names=out_names,
            channel_dim=channel_dim,
            normalizer=normalizer,
        )



import dataclasses
from typing import Literal

from fme.ace.models.makani.sfnonet import (
    SphericalFourierNeuralOperatorNet as MakaniSFNO,
)
from fme.ace.models.modulus.sfnonet import SphericalFourierNeuralOperatorNet
from fme.ace.registry.registry import ModuleConfig, ModuleSelector


# this is based on the call signature of SphericalFourierNeuralOperatorNet at
# https://github.com/NVIDIA/modulus/blob/b8e27c5c4ebc409e53adaba9832138743ede2785/modulus/models/sfno/sfnonet.py#L292  # noqa: E501
@ModuleSelector.register("SphericalFourierNeuralOperatorNet")
@dataclasses.dataclass
class SphericalFourierNeuralOperatorBuilder(ModuleConfig):
    """
    Configuration for the SFNO architecture used in FourCastNet-SFNO.
    """

    spectral_transform: str = "sht"
    filter_type: str = "linear"
    operator_type: str = "diagonal"
    scale_factor: int = 1
    residual_filter_factor: int = 1
    embed_dim: int = 256
    num_layers: int = 12
    hard_thresholding_fraction: float = 1.0
    normalization_layer: str = "instance_norm"
    use_mlp: bool = True
    activation_function: str = "gelu"
    encoder_layers: int = 1
    pos_embed: bool = True
    big_skip: bool = True
    rank: float = 1.0
    factorization: str | None = None
    separable: bool = False
    complex_network: bool = True
    complex_activation: str = "real"
    spectral_layers: int = 1
    checkpointing: int = 0
    data_grid: Literal["legendre-gauss", "equiangular"] = "legendre-gauss"

    def build(
        self,
        n_in_channels: int,
        n_out_channels: int,
        img_shape: tuple[int, int],
    ):
        sfno_net = SphericalFourierNeuralOperatorNet(
            params=self,
            in_chans=n_in_channels,
            out_chans=n_out_channels,
            img_shape=img_shape,
        )

        return sfno_net


@ModuleSelector.register("SFNO-v0.1.0")
@dataclasses.dataclass
class SFNO_V0_1_0(ModuleConfig):
    """
    Configuration for the SFNO architecture in modulus-makani version 0.1.0.
    """

    spectral_transform: str = "sht"
    filter_type: Literal["linear"] = "linear"
    operator_type: str = "dhconv"
    scale_factor: int = 16
    embed_dim: int = 256
    num_layers: int = 12
    repeat_layers: int = 1
    hard_thresholding_fraction: float = 1.0
    normalization_layer: str = "instance_norm"
    use_mlp: bool = True
    activation_function: str = "gelu"
    encoder_layers: int = 1
    pos_embed: Literal["none", "direct", "frequency"] = "direct"
    big_skip: bool = True
    rank: float = 1.0
    factorization: str | None = None
    separable: bool = False
    complex_activation: str = "real"
    spectral_layers: int = 1
    checkpointing: int = 0
    data_grid: Literal["legendre-gauss", "equiangular", "healpix"] = "legendre-gauss"

    def build(
        self,
        n_in_channels: int,
        n_out_channels: int,
        img_shape: tuple[int, int],
    ):
        return MakaniSFNO(
            inp_chans=n_in_channels,
            out_chans=n_out_channels,
            inp_shape=img_shape,
            out_shape=img_shape,
            model_grid_type=self.data_grid,
            spectral_transform=self.spectral_transform,
            filter_type=self.filter_type,
            operator_type=self.operator_type,
            scale_factor=self.scale_factor,
            embed_dim=self.embed_dim,
            num_layers=self.num_layers,
            repeat_layers=self.repeat_layers,
            hard_thresholding_fraction=self.hard_thresholding_fraction,
            normalization_layer=self.normalization_layer,
            use_mlp=self.use_mlp,
            activation_function=self.activation_function,
            encoder_layers=self.encoder_layers,
            pos_embed=self.pos_embed,
            big_skip=self.big_skip,
            rank=self.rank,
            factorization=self.factorization,
            separable=self.separable,
            complex_activation=self.complex_activation,
            spectral_layers=self.spectral_layers,
            checkpointing=self.checkpointing,
        )



import dataclasses
import os
import warnings
from collections.abc import Callable, Mapping, Sequence
from typing import Any

import torch

from fme.ace.aggregator import (
    InferenceEvaluatorAggregatorConfig,
    OneStepAggregatorConfig,
)
from fme.ace.aggregator.inference.main import InferenceEvaluatorAggregator
from fme.ace.data_loading.config import DataLoaderConfig
from fme.ace.data_loading.getters import get_gridded_data, get_inference_data
from fme.ace.data_loading.gridded_data import (
    ErrorInferenceData,
    GriddedData,
    InferenceGriddedData,
)
from fme.ace.data_loading.inference import InferenceDataLoaderConfig
from fme.ace.requirements import (
    DataRequirements,
    NullDataRequirements,
    PrognosticStateDataRequirements,
)
from fme.ace.stepper import ExistingStepperConfig, SingleModuleStepperConfig, Stepper
from fme.ace.stepper.single_module import StepperConfig
from fme.core.dataset.data_typing import VariableMetadata
from fme.core.dataset_info import DatasetInfo
from fme.core.distributed import Distributed
from fme.core.ema import EMAConfig, EMATracker
from fme.core.generics.trainer import EndOfBatchCallback, EndOfEpochCallback
from fme.core.logging_utils import LoggingConfig
from fme.core.optimization import Optimization, OptimizationConfig
from fme.core.typing_ import Slice, TensorDict, TensorMapping
from fme.core.weight_ops import CopyWeightsConfig


@dataclasses.dataclass
class WeatherEvaluationConfig:
    """
    Parameters:
        loader: configuration for the data loader used during weather evaluation
        n_forward_steps: number of forward steps to take
        forward_steps_in_memory: number of forward steps to take before
            re-reading data from disk
        epochs: epochs on which to run weather evaluation. By default runs
            weather evaluation every epoch.
        aggregator: configuration of weather evaluation aggregator.
    """

    loader: InferenceDataLoaderConfig
    n_forward_steps: int = 2
    forward_steps_in_memory: int = 2
    epochs: Slice = dataclasses.field(default_factory=lambda: Slice())
    aggregator: InferenceEvaluatorAggregatorConfig = dataclasses.field(
        default_factory=lambda: InferenceEvaluatorAggregatorConfig(
            log_global_mean_time_series=False, log_global_mean_norm_time_series=False
        )
    )

    def __post_init__(self):
        dist = Distributed.get_instance()
        if self.loader.start_indices.n_initial_conditions % dist.world_size != 0:
            raise ValueError(
                "Number of inference initial conditions must be divisible by the "
                "number of parallel workers, got "
                f"{self.loader.start_indices.n_initial_conditions} and "
                f"{dist.world_size}."
            )
        if (
            self.aggregator.log_global_mean_time_series
            or self.aggregator.log_global_mean_norm_time_series
        ):
            # Both of log_global_mean_time_series and
            # log_global_mean_norm_time_series must be False for inline inference.
            self.aggregator.log_global_mean_time_series = False
            self.aggregator.log_global_mean_norm_time_series = False

    def get_inference_data(
        self,
        window_requirements: DataRequirements,
        initial_condition: PrognosticStateDataRequirements,
    ) -> InferenceGriddedData:
        return get_inference_data(
            config=self.loader,
            total_forward_steps=self.n_forward_steps,
            window_requirements=window_requirements,
            initial_condition=initial_condition,
        )


@dataclasses.dataclass
class InlineInferenceConfig:
    """
    Parameters:
        loader: configuration for the data loader used during inference
        n_forward_steps: number of forward steps to take
        forward_steps_in_memory: number of forward steps to take before
            re-reading data from disk
        epochs: epochs on which to run inference. By default runs inference every epoch.
        aggregator: configuration of inline inference aggregator.
    """

    loader: InferenceDataLoaderConfig
    n_forward_steps: int = 2
    forward_steps_in_memory: int = 2
    epochs: Slice = dataclasses.field(default_factory=lambda: Slice())
    aggregator: InferenceEvaluatorAggregatorConfig = dataclasses.field(
        default_factory=lambda: InferenceEvaluatorAggregatorConfig(
            log_global_mean_time_series=False, log_global_mean_norm_time_series=False
        )
    )

    def __post_init__(self):
        dist = Distributed.get_instance()
        if self.loader.start_indices.n_initial_conditions % dist.world_size != 0:
            raise ValueError(
                "Number of inference initial conditions must be divisible by the "
                "number of parallel workers, got "
                f"{self.loader.start_indices.n_initial_conditions} and "
                f"{dist.world_size}."
            )
        if (
            self.aggregator.log_global_mean_time_series
            or self.aggregator.log_global_mean_norm_time_series
        ):
            # Both of log_global_mean_time_series and
            # log_global_mean_norm_time_series must be False for inline inference.
            self.aggregator.log_global_mean_time_series = False
            self.aggregator.log_global_mean_norm_time_series = False

    def get_inference_data(
        self,
        window_requirements: DataRequirements,
        initial_condition: PrognosticStateDataRequirements,
    ) -> InferenceGriddedData:
        return get_inference_data(
            config=self.loader,
            total_forward_steps=self.n_forward_steps,
            window_requirements=window_requirements,
            initial_condition=initial_condition,
        )


@dataclasses.dataclass
class TrainConfig:
    """
    Configuration for training a model.

    Arguments:
        train_loader: Configuration for the training data loader.
        validation_loader: Configuration for the validation data loader.
        stepper: Configuration for the stepper. SingleModuleStepperConfig is
            deprecated and will be removed in a future version. Use StepperConfig
            instead.
        optimization: Configuration for the optimization.
        logging: Configuration for logging.
        max_epochs: Total number of epochs to train for.
        save_checkpoint: Whether to save checkpoints.
        experiment_dir: Directory where checkpoints and logs are saved.
        inference: Configuration for inline inference.
            If None, no inline inference is run,
            and no "best_inline_inference" checkpoint will be saved.
        weather_evaluation: Configuration for weather evaluation.
            If None, no weather evaluation is run. Weather evaluation is not
            used to select checkpoints, but is used to provide metrics.
        n_forward_steps: Number of forward steps to take gradient over.
        copy_weights_after_batch: Configuration for copying weights from the
            base model to the training model after each batch.
        ema: Configuration for exponential moving average of model weights.
        validate_using_ema: Whether to validate and perform inference using
            the EMA model.
        checkpoint_save_epochs: How often to save epoch-based checkpoints,
            if save_checkpoint is True. If None, checkpoints are only saved
            for the most recent epoch
            (and the best epochs if validate_using_ema == False).
        ema_checkpoint_save_epochs: How often to save epoch-based EMA checkpoints,
            if save_checkpoint is True. If None, EMA checkpoints are only saved
            for the most recent epoch
            (and the best epochs if validate_using_ema == True).
        log_train_every_n_batches: How often to log batch_loss during training.
        segment_epochs: Exit after training for at most this many epochs
            in current job, without exceeding `max_epochs`. Use this if training
            must be run in segments, e.g. due to wall clock limit.
        save_per_epoch_diagnostics: Whether to save per-epoch diagnostics from
            training, validation and inline inference aggregators.
        validation_aggregator: Configuration for the validation aggregator.
        evaluate_before_training: Whether to run validation and inline inference before
            any training is done.
    """

    train_loader: DataLoaderConfig
    validation_loader: DataLoaderConfig
    stepper: SingleModuleStepperConfig | ExistingStepperConfig | StepperConfig
    optimization: OptimizationConfig
    logging: LoggingConfig
    max_epochs: int
    save_checkpoint: bool
    experiment_dir: str
    inference: InlineInferenceConfig | None
    n_forward_steps: int
    copy_weights_after_batch: list[CopyWeightsConfig] = dataclasses.field(
        default_factory=list
    )
    ema: EMAConfig = dataclasses.field(default_factory=lambda: EMAConfig())
    weather_evaluation: WeatherEvaluationConfig | None = None
    validate_using_ema: bool = False
    checkpoint_save_epochs: Slice | None = None
    ema_checkpoint_save_epochs: Slice | None = None
    log_train_every_n_batches: int = 100
    segment_epochs: int | None = None
    save_per_epoch_diagnostics: bool = False
    validation_aggregator: OneStepAggregatorConfig = dataclasses.field(
        default_factory=lambda: OneStepAggregatorConfig()
    )
    evaluate_before_training: bool = False

    def __post_init__(self):
        if isinstance(self.stepper, SingleModuleStepperConfig):
            warnings.warn(
                "SingleModuleStepperConfig is deprecated. Use StepperConfig instead.",
                DeprecationWarning,
            )

    @property
    def inference_n_forward_steps(self) -> int:
        if self.inference is None:
            return 0
        return self.inference.n_forward_steps

    @property
    def inference_aggregator(self) -> InferenceEvaluatorAggregatorConfig | None:
        if self.inference is None:
            return None
        return self.inference.aggregator

    @property
    def checkpoint_dir(self) -> str:
        """
        The directory where checkpoints are saved.
        """
        return os.path.join(self.experiment_dir, "training_checkpoints")

    @property
    def output_dir(self) -> str:
        """
        The directory where output files are saved.
        """
        return os.path.join(self.experiment_dir, "output")

    def get_inference_epochs(self) -> list[int]:
        if self.inference is None:
            return []
        start_epoch = 0 if self.evaluate_before_training else 1
        all_epochs = list(range(start_epoch, self.max_epochs + 1))
        return all_epochs[self.inference.epochs.slice]


class TrainBuilders:
    def __init__(self, config: TrainConfig):
        self.config = config

    def _get_train_window_data_requirements(self) -> DataRequirements:
        return self.config.stepper.get_evaluation_window_data_requirements(
            self.config.n_forward_steps
        )

    def _get_evaluation_window_data_requirements(self) -> DataRequirements:
        if self.config.inference is None:
            return NullDataRequirements
        return self.config.stepper.get_evaluation_window_data_requirements(
            self.config.inference.forward_steps_in_memory
        )

    def _get_initial_condition_data_requirements(
        self,
    ) -> PrognosticStateDataRequirements:
        return self.config.stepper.get_prognostic_state_data_requirements()

    def get_train_data(self) -> GriddedData:
        data_requirements = self._get_train_window_data_requirements()
        return get_gridded_data(
            self.config.train_loader,
            requirements=data_requirements,
            train=True,
        )

    def get_validation_data(self) -> GriddedData:
        data_requirements = self._get_train_window_data_requirements()
        return get_gridded_data(
            self.config.validation_loader,
            requirements=data_requirements,
            train=False,
        )

    def get_evaluation_inference_data(
        self,
    ) -> InferenceGriddedData:
        if self.config.inference is None:
            return ErrorInferenceData()  # type: ignore
        else:
            return self.config.inference.get_inference_data(
                window_requirements=self._get_evaluation_window_data_requirements(),
                initial_condition=self._get_initial_condition_data_requirements(),
            )

    def get_optimization(self, modules: torch.nn.ModuleList) -> Optimization:
        return self.config.optimization.build(modules, self.config.max_epochs)

    def get_stepper(
        self,
        dataset_info: DatasetInfo,
    ) -> Stepper:
        return self.config.stepper.get_stepper(
            dataset_info=dataset_info,
        )

    def get_ema(self, modules) -> EMATracker:
        return self.config.ema.build(modules)

    def get_end_of_batch_ops(
        self,
        modules: list[torch.nn.Module],
        base_weights: list[Mapping[str, Any]] | None,
    ) -> EndOfBatchCallback:
        if base_weights is not None:

            def copy_after_batch():
                for module, copy_config in zip(
                    modules, self.config.copy_weights_after_batch
                ):
                    copy_config.apply(weights=base_weights, modules=[module])
                return

            return copy_after_batch
        return lambda: None

    def get_end_of_epoch_callback(
        self,
        inference_one_epoch: Callable[
            [InferenceGriddedData, InferenceEvaluatorAggregator, str, int],
            Mapping[str, Any],
        ],
        normalize: Callable[[TensorMapping], TensorDict],
        output_dir: str,
        variable_metadata: Mapping[str, VariableMetadata],
        channel_mean_names: Sequence[str],
        save_diagnostics: bool,
        n_ic_timesteps: int,
    ) -> EndOfEpochCallback:
        if self.config.weather_evaluation is not None:
            data = self.config.weather_evaluation.get_inference_data(
                window_requirements=self._get_evaluation_window_data_requirements(),
                initial_condition=self._get_initial_condition_data_requirements(),
            )
            dataset_info = data.dataset_info.update_variable_metadata(variable_metadata)
            aggregator = self.config.weather_evaluation.aggregator.build(
                dataset_info=dataset_info,
                n_timesteps=self.config.weather_evaluation.n_forward_steps
                + n_ic_timesteps,
                initial_time=data.initial_time,
                normalize=normalize,
                output_dir=output_dir,
                record_step_20=self.config.weather_evaluation.n_forward_steps >= 20,
                channel_mean_names=channel_mean_names,
                save_diagnostics=save_diagnostics,
            )

            def end_of_epoch_ops(epoch: int) -> Mapping[str, Any]:
                if self.config.weather_evaluation is not None:
                    if self.config.weather_evaluation.epochs.contains(epoch):
                        return inference_one_epoch(
                            data,
                            aggregator,
                            "weather_eval",
                            epoch,
                        )
                return {}

            return end_of_epoch_ops

        return lambda epoch: {}



#!/usr/bin/env python
#
# fme documentation build configuration file, created by
# sphinx-quickstart on Fri Jun  9 13:47:02 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another
# directory, add these directories to sys.path here. If the directory is
# relative to the documentation root, use os.path.abspath to make it
# absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath(".."))

import fme  # noqa
import fme.core.registry

# -- General configuration ---------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Fetch the dynamic data
module_types = fme.core.registry.ModuleSelector.get_available_types()

# Create a dynamic rst snippet that can be included in your documentation
rst_snippet = f".. code-block:: text\n\n   {module_types}"

# Save the rst snippet to a file that can be included in your Sphinx documentation
with open("available_modules.rst", "w") as f:
    f.write(rst_snippet)


# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx_autodoc_typehints",
]
autodoc_typehints = "description"

# Intersphinx configuration to link to other projects' documentation
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "ACE"
copyright = "2024, Allen Institute for Artificial Intelligence"
author = "Allen Institute for Artificial Intelligence"

# The version info for the project you're documenting, acts as replacement
# for |version| and |release|, also used in various other places throughout
# the built documents.
#
# The short X.Y version.
version = fme.__version__
# The full version, including alpha/beta/rc tags.
release = fme.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# Include default values when documenting parameter types.
typehints_defaults = "comma"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output -------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"
html_title = f"fme v{fme.__version__}"

# Theme options are theme-specific and customize the look and feel of a
# theme further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    "light_logo": "Ai2_icon_pink_padding_RGB.png",
    "dark_logo": "Ai2_icon_pink_padding_RGB.png",
    "footer_icons": [
        {
            "name": "GitHub",
            "url": "https://github.com/ai2cm/ace",
            "html": """
                <svg stroke="currentColor" fill="currentColor" stroke-width="0" viewBox="0 0 16 16">
                    <path fill-rule="evenodd" d="M8 0C3.58 0 0 3.58 0 8c0 3.54 2.29 6.53 5.47 7.59.4.07.55-.17.55-.38 0-.19-.01-.82-.01-1.49-2.01.37-2.53-.49-2.69-.94-.09-.23-.48-.94-.82-1.13-.28-.15-.68-.52-.01-.53.63-.01 1.08.58 1.23.82.72 1.21 1.87.87 2.33.66.07-.52.28-.87.51-1.07-1.78-.2-3.64-.89-3.64-3.95 0-.87.31-1.59.82-2.15-.08-.2-.36-1.02.08-2.12 0 0 .67-.21 2.2.82.64-.18 1.32-.27 2-.27.68 0 1.36.09 2 .27 1.53-1.04 2.2-.82 2.2-.82.44 1.1.16 1.92.08 2.12.51.56.82 1.27.82 2.15 0 3.07-1.87 3.75-3.65 3.95.29.25.54.73.54 1.48 0 1.07-.01 1.93-.01 2.2 0 .21.15.46.55.38A8.013 8.013 0 0 0 16 8c0-4.42-3.58-8-8-8z"></path>
                </svg>
            """,  # noqa: E501
            "class": "",
        },
    ],
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_css_files = ["custom.css"]

html_favicon = "_static/Ai2_icon_pink_RGB.png"


# -- Options for HTMLHelp output ---------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "fmedoc"


# -- Options for LaTeX output ------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "fme.tex",
        "ACE Documentation",
        "Allen Institute for Artificial Intelligence",
        "manual",
    ),
]


# -- Options for manual page output ------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "fme", "ACE Documentation", [author], 1)]


# -- Options for Texinfo output ----------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "fme",
        "ACE Documentation",
        author,
        "fme",
        "One line description of project.",
        "Miscellaneous",
    ),
]


# -- Options for doctest -----------------------------------------------
import doctest

IGNORE_OUTPUT = doctest.register_optionflag("IGNORE_OUTPUT")

OutputChecker = doctest.OutputChecker


class CustomOutputChecker(OutputChecker):
    def check_output(self, want, got, optionflags):
        if IGNORE_OUTPUT & optionflags:
            return True
        return OutputChecker.check_output(self, want, got, optionflags)


doctest.OutputChecker = CustomOutputChecker

doctest_global_setup = """
import fme
"""



import torch

from fme.ace.aggregator.inference.reduced import MeanAggregator
from fme.ace.aggregator.inference.time_mean import TimeMeanEvaluatorAggregator
from fme.core.device import get_device
from fme.core.gridded_ops import LatLonOperations
from fme.core.testing import mock_distributed


def test_mean_metrics_call_distributed():
    """
    All mean metrics should be reduced across processes using Distributed.

    This tests that functionality by modifying the Distributed singleton.
    """
    with mock_distributed(-1.0):
        data_a = torch.ones([2, 3, 4, 4], device=get_device())
        area_weights = torch.ones(1).to(get_device())
        agg = MeanAggregator(
            LatLonOperations(area_weights), target="denorm", n_timesteps=3
        )
        sample_data = {"a": data_a}
        agg.record_batch(sample_data, sample_data, sample_data, sample_data)
        logs = agg.get_logs(label="metrics")
        table = logs["metrics/series"]
        # assert all data past the first column in the WandB table is -1
        assert all([all(item == -1 for item in row[1][1:]) for row in table.iterrows()])


def test_time_mean_metrics_call_distributed():
    """
    All time-mean metrics should be reduced across processes using Distributed.

    This tests that functionality by modifying the Distributed singleton.
    """
    torch.manual_seed(0)
    with mock_distributed(0.0) as mock:
        area_weights = torch.ones(1).to(get_device())
        agg = TimeMeanEvaluatorAggregator(
            LatLonOperations(area_weights), horizontal_dims=["lat", "lon"]
        )
        target_data = {"a": torch.ones([2, 3, 4, 4], device=get_device())}
        gen_data = {"a": torch.randn([2, 3, 4, 4], device=get_device())}
        agg.record_batch(
            target_data=target_data,
            gen_data=gen_data,
            target_data_norm=target_data,
            gen_data_norm=gen_data,
        )
        logs = agg.get_logs(label="metrics")
        # the reduction happens on the time-means, so the gen and target data should
        # be filled identically and all errors will be zero, even though we gave them
        # different data above.
        assert logs["metrics/rmse/a"] == 0.0
        assert logs["metrics/bias/a"] == 0.0
        assert mock.reduce_called



import logging
import os
from collections.abc import Callable

import torch.distributed
from torch.nn import SyncBatchNorm
from torch.nn.functional import pad
from torch.nn.parallel import DistributedDataParallel

from fme.core.device import get_device, using_gpu

logger = logging.getLogger(__name__)


class DummyWrapper(torch.nn.Module):
    """
    Wrapper class for a single pytorch module, which does nothing.

    Exists so we have an identical module structure to the case where we use
    a DistributedDataParallel wrapper.
    """

    def __init__(self, module):
        super().__init__()
        self.module = module

    def forward(self, *args, **kwargs):
        return self.module(*args, **kwargs)


class Distributed:
    """
    A class to represent the distributed concerns for FME training.

    This should generally be initialized first, before any pytorch objects.
    This is important because it sets global variables such as the CUDA
    device for the local rank, which is used when initializing pytorch objects.

    This class uses the
    [Singleton pattern](https://en.wikipedia.org/wiki/Singleton_pattern) and should
    be initialized through get_instance. This pattern allows easy access to global
    variables without having to pass them around, and lets us put the initialization
    for this global state in the same place as the routines that use it.

    Parameters:
        world_size: The number of processes in the distributed training job.
        rank: The global rank of the current process.
        local_rank: The node-local rank of the current process.
    """

    @classmethod
    def get_instance(cls) -> "Distributed":
        """
        Get the singleton instance of the Distributed class.
        """
        global singleton
        if singleton is None:
            singleton = cls()
        return singleton

    def __init__(self):
        if torch.distributed.is_available() and not torch.distributed.is_initialized():
            self._distributed = self._init_distributed()
        else:
            self._distributed = False

    def _init_distributed(self):
        if "RANK" in os.environ:  # we were executed with torchrun
            if using_gpu():
                torch.distributed.init_process_group(
                    backend="nccl", init_method="env://"
                )
            else:
                torch.distributed.init_process_group(
                    backend="gloo", init_method="env://"
                )
            self.world_size = torch.distributed.get_world_size()
            self.local_rank = int(os.environ["LOCAL_RANK"])
            self.rank = torch.distributed.get_rank()
            if using_gpu():
                torch.cuda.set_device(self.local_rank)
            distributed = True
        else:
            self.world_size = 1
            self.rank = 0
            self.local_rank = 0
            distributed = False
        return distributed

    def local_batch_size(self, batch_size: int) -> int:
        """
        Get the local batch size for the current process.
        """
        return batch_size // self.world_size

    def reduce_mean(self, tensor: torch.Tensor) -> torch.Tensor:
        """
        Reduce a tensor representing a mean across all processes.

        Whether the tensor represents a mean is important because to reduce a mean,
        we must divide by the number of processes. To reduce a sum, we must not.

        Modifies the input tensor in-place as a side effect.
        """
        if self._distributed:
            torch.distributed.all_reduce(tensor)
        return tensor / self.world_size

    def reduce_sum(self, tensor: torch.Tensor) -> torch.Tensor:
        """
        Reduce a tensor representing a sum across all processes.

        Whether the tensor represents a mean is important because to reduce a mean,
        we must divide by the number of processes. To reduce a sum, we must not.

        Modifies the input tensor in-place as a side effect.
        """
        if self._distributed:
            torch.distributed.all_reduce(tensor)
        return tensor

    def reduce_min(self, tensor: torch.Tensor) -> torch.Tensor:
        """
        Reduce a tensor representing a min across all processes.

        Modifies the input tensor in-place as a side effect.
        """
        if self._distributed:
            torch.distributed.all_reduce(tensor, op=torch.distributed.ReduceOp.MIN)
        return tensor

    def reduce_max(self, tensor: torch.Tensor) -> torch.Tensor:
        """
        Reduce a tensor representing a max across all processes.

        Modifies the input tensor in-place as a side effect.
        """
        if self._distributed:
            torch.distributed.all_reduce(tensor, op=torch.distributed.ReduceOp.MAX)
        return tensor

    def gather(self, tensor: torch.Tensor) -> list[torch.Tensor] | None:
        """
        Gather a tensor from all processes to the root process.

        Note: tensor shape is assumed to be equal across all processes; data will
            reshaped/filled/dropped to coerce non-root tensors to the shape
            of the root tensor if not. To avoid this behavior, use
            "gather_irregular" instead.

        Args:
            tensor: The tensor to gather.

        Returns:
            A list of tensors, where the i-th element is the tensor
                from the i-th process.
        """
        gather_list: list[torch.Tensor] | None = None
        if self.rank == 0:
            gather_list = [tensor] + [
                torch.empty_like(tensor) for _ in range(self.world_size - 1)
            ]
        if self._distributed:
            torch.distributed.gather(tensor, gather_list)
        return gather_list

    def gather_irregular(
        self,
        tensor: torch.Tensor,
    ) -> list[torch.Tensor] | None:
        """
        Gather a tensor from all processes to the root process. The rank tensors
        may have diferent dimension lengths, but must have the same number of
        dimensions.

        Args:
            tensor: The tensor to gather.

        Returns:
            A list of tensors of consistent shape, where the i-th element is the tensor
                from the i-th process.
        """
        return gather_irregular(
            tensor,
            self.reduce_max,
            self.gather,
            is_distributed=self.is_distributed(),
        )

    def is_root(self) -> bool:
        """
        Returns True if this process is the root process.
        """
        return self.rank == 0

    def is_distributed(self) -> bool:
        """
        Returns True if this process is running in a distributed context
        with more than 1 worker.
        """
        return self._distributed and self.world_size > 1

    def wrap_module(self, module: torch.nn.Module) -> torch.nn.Module:
        """
        Wrap a model with DistributedDataParallel if running in a distributed context.
        """
        if self.is_distributed() and any(p.requires_grad for p in module.parameters()):
            if using_gpu():
                device_ids = [self.local_rank]
                output_device = [self.local_rank]
            else:
                device_ids = None
                output_device = None
            return DistributedDataParallel(
                SyncBatchNorm.convert_sync_batchnorm(module),
                device_ids=device_ids,
                output_device=output_device,
            )
        else:
            return DummyWrapper(module)

    def barrier(self):
        """
        Wait for all processes to reach this point.
        """
        if self._distributed:
            logger.debug(f"Barrier on rank {self.rank}")
            torch.distributed.barrier()


singleton: Distributed | None = None


def gather_irregular(
    tensor: torch.Tensor,
    reduce_max: Callable[[torch.Tensor], torch.tensor],
    gather: Callable[[torch.Tensor], list[torch.Tensor] | None],
    is_distributed: bool = False,
    fill_value: float | int = 0.0,
) -> list[torch.Tensor] | None:
    """
    Gather a tensor from all processes to the root process. The rank tensors
    may have diferent dimension lengths, but must have the same number of dimensions.
    temporarily padded with `fill_value` where its dimension length is smaller than the
    maxmimum dimension length, but the padding is removed prior to returning the
    gathered tensors.

    Args:
        tensor: The tensor to gather.
        reduce_max: The reduction function to use for each dimension length.
        gather: The gather function to use.
        is_distributed: Whether the current process is distributed.
        fill_value: The value to fill each tensor with.

    Returns:
        A list of tensors, where the i-th element is the tensor from the i-th process.
    """
    output_tensor_size = []
    tensor_size = list(tensor.size())
    for dim_len in tensor_size:
        if is_distributed:
            dimension_length = torch.tensor(
                dim_len, dtype=torch.int32, device=get_device()
            )
            reduce_max(dimension_length)
            output_tensor_size.append(int(dimension_length.item()))
        else:
            output_tensor_size.append(int(dim_len))
    dimension_difference = [
        output - input for output, input in zip(output_tensor_size, tensor_size)
    ]
    regular_tensor = pad_tensor_at_end(tensor, dimension_difference, fill_value)
    gathered_regular_tensors = gather(regular_tensor)
    gathered_dimension_differences = gather(
        torch.tensor(dimension_difference, device=get_device())
    )
    if gathered_regular_tensors is None or gathered_dimension_differences is None:
        return None
    else:
        return [
            unpad_tensor_at_end(regular_tensor, dimension_difference)
            for regular_tensor, dimension_difference in zip(
                gathered_regular_tensors, gathered_dimension_differences
            )
        ]


def pad_tensor_at_end(
    tensor: torch.Tensor,
    dimension_difference: list[int],
    fill_value: float | int = 0.0,
):
    """Pad tensor by specified amount at end of each dimension.
    Note that `pad` format is in reverse dimension order.

    Args:
        tensor: The tensor to pad
        dimension_difference: The amount to pad each dimension
        fill_value: The value to fill the padding with

    Returns:
        The padded tensor
    """
    assert len(dimension_difference) == len(tensor.size()), "Dimension mismatch"
    pad_dimensions = tuple(
        [
            val
            for pair in zip(
                [0 for _ in tensor.size()],
                [diff for diff in dimension_difference[::-1]],
            )
            for val in pair
        ]
    )
    padded_tensor = pad(tensor, pad_dimensions, mode="constant", value=fill_value).to(
        get_device()
    )
    return padded_tensor


def unpad_tensor_at_end(
    tensor: torch.Tensor, dimension_difference: torch.Tensor
) -> torch.Tensor:
    """Remove padding from tensor.

    Args:
        tensor: The tensor to remove padding from
        dimension_difference: The amount of padding to remove from each dimension

    Returns:
        The tensor with padding removed
    """
    assert len(dimension_difference) == len(tensor.size()), "Dimension mismatch"
    slice_dimensions = tuple(
        [
            slice(0, tensor.size()[i] - dimension_difference[i])
            for i in range(len(tensor.size()))
        ]
    )
    return tensor[slice_dimensions]



import abc
import contextlib

import torch
from torch import nn

from fme.core.typing_ import TensorDict, TensorMapping


class OptimizationABC(abc.ABC):
    @contextlib.contextmanager
    @abc.abstractmethod
    def autocast(self): ...

    @property
    @abc.abstractmethod
    def learning_rate(self) -> float: ...

    @abc.abstractmethod
    def set_mode(self, modules: nn.ModuleList):
        """
        Sets the mode of the module to train.
        """
        ...

    @abc.abstractmethod
    def step_scheduler(self, valid_loss: float):
        """
        Step the scheduler.

        Args:
            valid_loss: The validation loss. Used in schedulers which change the
                learning rate based on whether the validation loss is decreasing.
        """
        ...

    @abc.abstractmethod
    def detach_if_using_gradient_accumulation(self, state: TensorMapping) -> TensorDict:
        """
        Detaches the state if using gradient accumulation.
        """
        ...

    @abc.abstractmethod
    def accumulate_loss(self, loss: torch.Tensor):
        """
        Accumulate the loss.

        In order to support gradient accumulation, loss values accumulated after
        a call to `accumulate_loss` must not depend on any parameter interactions that
        occurred before the call. For example, if this is called at the end of a model
        step, later steps must have their graph detached from previous timesteps. This
        can be done only when gradient accumulation is enabled using the
        `detach_if_using_gradient_accumulation` method on this object.
        """
        ...

    @abc.abstractmethod
    def checkpoint(self, module: nn.Module, step: int) -> nn.Module:
        """
        Applies activation checkpointing to the module if configured to do so,
        otherwise returns the module unchanged.

        Args:
            module: The module to checkpoint.
            step: The current step number.
        """
        ...

    @abc.abstractmethod
    def get_accumulated_loss(self) -> torch.Tensor:
        """
        Get the accumulated loss.
        """
        ...

    @abc.abstractmethod
    def step_weights(self):
        """
        Step the weights.

        Resets the accumulated loss to zero.
        """
        ...

    @abc.abstractmethod
    def get_state(self):
        """
        Returns state as a serializable data structure.
        """
        ...

    @abc.abstractmethod
    def load_state(self, state):
        """
        Loads state from a serializable data structure.
        """
        ...



import datetime
from collections.abc import Mapping

import cftime
import numpy as np
import pytest
import xarray as xr

from fme.ace.data_loading.config import ConcatDatasetConfig, DataLoaderConfig
from fme.ace.data_loading.getters import get_gridded_data
from fme.ace.requirements import DataRequirements
from fme.core.dataset.config import XarrayDataConfig
from fme.core.dataset.data_typing import VariableMetadata


def _coord_value(name, size):
    # xarray data loader requires time to be cftime.datetime object
    if name == "time":
        return [
            cftime.DatetimeProlepticGregorian(2000, 1, 1) + datetime.timedelta(hours=i)
            for i in range(size)
        ]
    else:
        return np.arange(size, dtype=np.float32)


def _save_netcdf(
    filename,
    variable_metadata: Mapping[str, VariableMetadata | None],
    num_members=1,
    dim_sizes=None,
):
    if dim_sizes is None:
        dim_sizes = {"time": 3, "lat": 16, "lon": 32}
    data_vars = {}
    for name in variable_metadata:
        data = np.random.randn(*list(dim_sizes.values()))
        if len(dim_sizes) > 0:
            data = data.astype(np.float32)
        item_metadata = variable_metadata[name]
        if item_metadata is None:
            attrs = {}
        else:
            attrs = {
                "units": item_metadata.units,
                "long_name": item_metadata.long_name,
            }
        data_vars[name] = xr.DataArray(data, dims=list(dim_sizes), attrs=attrs)
    coords = {
        dim_name: xr.DataArray(
            _coord_value(dim_name, size),
            dims=(dim_name,),
        )
        for dim_name, size in dim_sizes.items()
    }

    for i in range(7):
        data_vars[f"ak_{i}"] = float(i)
        data_vars[f"bk_{i}"] = float(i + 1)

    ds = xr.Dataset(data_vars=data_vars, coords=coords)
    ds.to_netcdf(filename, unlimited_dims=["time"], format="NETCDF4_CLASSIC")


@pytest.mark.parametrize("n_ensemble_members", [1, 2])
@pytest.mark.parametrize(
    "variable_metadata",
    [
        pytest.param(
            {"bar": None},
            id="one_var_no_metadata",
        ),
        pytest.param(
            {"bar": VariableMetadata("km", "bar_long_name")},
            id="one_var_metadata",
        ),
        pytest.param(
            {"foo": VariableMetadata("m", "foo_long_name"), "bar": None},
            id="two_vars_one_metadata",
        ),
        pytest.param(
            {
                "foo": VariableMetadata("m", "foo_long_name"),
                "bar": VariableMetadata("km", "bar_long_name"),
            },
            id="two_vars_two_metadata",
        ),
    ],
)
def test_metadata(tmp_path, variable_metadata, n_ensemble_members):
    paths = []
    for i in range(n_ensemble_members):
        path = tmp_path / f"ic{i}"
        path.mkdir(exist_ok=True)
        paths.append(path)
        _save_netcdf(path / "data.nc", variable_metadata)

    config = DataLoaderConfig(
        dataset=ConcatDatasetConfig(
            concat=[XarrayDataConfig(data_path=str(path)) for path in paths]
        ),
        batch_size=1,
        num_data_workers=0,
    )
    var_names = list(variable_metadata.keys())
    requirements = DataRequirements(names=var_names, n_timesteps=2)
    data = get_gridded_data(config=config, train=True, requirements=requirements)
    target_metadata = {
        name: variable_metadata[name]
        for name in variable_metadata
        if variable_metadata[name] is not None
    }
    assert data.variable_metadata == target_metadata  # type: ignore



# flake8: noqa
# Copied from https://github.com/NVIDIA/modulus/commit/89a6091bd21edce7be4e0539cbd91507004faf08
# Copyright (c) 2023 - 2024 NVIDIA CORPORATION & AFFILIATES.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This file contains padding and convolution classes to perform according operations on the twelve faces of the HEALPix.


         HEALPix                              Face order                 3D array representation
                                                                            -----------------
--------------------------               //\\  //\\  //\\  //\\             |   |   |   |   |
|| 0  |  1  |  2  |  3  ||              //  \\//  \\//  \\//  \\            |0  |1  |2  |3  |
|\\  //\\  //\\  //\\  //|             /\\0 //\\1 //\\2 //\\3 //            -----------------
| \\//  \\//  \\//  \\// |            // \\//  \\//  \\//  \\//             |   |   |   |   |
|4//\\5 //\\6 //\\7 //\\4|            \\4//\\5 //\\6 //\\7 //\\             |4  |5  |6  |7  |
|//  \\//  \\//  \\//  \\|             \\/  \\//  \\//  \\//  \\            -----------------
|| 8  |  9  |  10 |  11  |              \\8 //\\9 //\\10//\\11//            |   |   |   |   |
--------------------------               \\//  \\//  \\//  \\//             |8  |9  |10 |11 |
                                                                            -----------------
                                    "\\" are top and bottom, whereas
                                    "//" are left and right borders


Details on the HEALPix can be found at https://iopscience.iop.org/article/10.1086/427976

"""

import logging
import sys

import torch as th
import torch.nn as nn


class HEALPixFoldFaces(nn.Module):
    """Class that folds the faces of a HealPIX tensor"""

    def __init__(self, enable_nhwc: bool = False):
        """
        Args:
            enable_nhwc: Use nhwc format instead of nchw format
        """
        super().__init__()
        self.enable_nhwc = enable_nhwc

    def forward(self, tensor: th.Tensor) -> th.Tensor:
        """
        Forward pass that folds a HEALPix tensor
        [B, F, C, H, W] -> [B*F, C, H, W]

        Args:
            tensor: The tensor to fold

        Returns:
            th.Tensor: the folded tensor

        """
        N, F, C, H, W = tensor.shape
        tensor = th.reshape(tensor, shape=(N * F, C, H, W))

        if self.enable_nhwc:
            tensor = tensor.to(memory_format=th.channels_last)

        return tensor


class HEALPixUnfoldFaces(nn.Module):
    """Class that unfolds the faces of a HealPIX tensor"""

    def __init__(self, num_faces=12, enable_nhwc=False):
        """
        Args:
            num_faces: The number of faces on the grid, default 12
            enable_nhwc: If nhwc format is being used, default False
        """
        super().__init__()
        self.num_faces = num_faces
        self.enable_nhwc = enable_nhwc

    def forward(self, tensor: th.Tensor) -> th.Tensor:
        """
        Forward pass that unfolds a HEALPix tensor
        [B*F, C, H, W] -> [B, F, C, H, W]

        Args:
            tensor: the tensor to unfold

        Returns:
            The unfolded tensor

        """
        NF, C, H, W = tensor.shape
        N = int(NF / self.num_faces)
        tensor = th.reshape(tensor, shape=(N, self.num_faces, C, H, W))

        return tensor


class HEALPixPadding(nn.Module):
    """
    Padding layer for data on a HEALPix sphere. The requirements for using this layer are as follows:
    - The last three dimensions are (face=12, height, width)
    - The first four indices in the faces dimension [0, 1, 2, 3] are the faces on the northern hemisphere
    - The second four indices in the faces dimension [4, 5, 6, 7] are the faces on the equator
    - The last four indices in the faces dimension [8, 9, 10, 11] are the faces on the southern hemisphere

    Orientation and arrangement of the HEALPix faces are outlined above.
    """

    def __init__(self, padding: int, enable_nhwc: bool = False):
        """
        Args:
            padding: The padding size
            enable_nhwc: If nhwc format is being used, default False
        """
        super().__init__()
        self.p = padding
        self.d = [-2, -1]
        self.enable_nhwc = enable_nhwc
        if not isinstance(padding, int) or padding < 1:
            raise ValueError(
                f"invalid value for 'padding', expected int > 0 but got {padding}"
            )

        self.fold = HEALPixFoldFaces(enable_nhwc=self.enable_nhwc)
        self.unfold = HEALPixUnfoldFaces(num_faces=12, enable_nhwc=self.enable_nhwc)

    def forward(self, data: th.Tensor) -> th.Tensor:
        """
        Pad each face consistently with its according neighbors in the HEALPix (see ordering and neighborhoods above).
        Assumes the Tensor is folded

        Args:
        data: The input tensor of shape [..., F, H, W] where each face is to be padded in its HPX context

        Returns:
            The padded tensor where each face's height and width are increased by 2*p
        """
        try:
            th.cuda.nvtx.range_push("HEALPixPadding:forward")
        except RuntimeError:
            # Don't bother with NVTX profiling if it's not available
            pass

        # unfold faces from batch dim
        data = self.unfold(data)

        # Extract the twelve faces (as views of the original tensors)
        f00, f01, f02, f03, f04, f05, f06, f07, f08, f09, f10, f11 = [
            th.squeeze(x, dim=1)
            for x in th.split(tensor=data, split_size_or_sections=1, dim=1)
        ]

        # Assemble the four padded faces on the northern hemisphere
        p00 = self.pn(
            c=f00, t=f01, tl=f02, lft=f03, bl=f03, b=f04, br=f08, rgt=f05, tr=f01
        )
        p01 = self.pn(
            c=f01, t=f02, tl=f03, lft=f00, bl=f00, b=f05, br=f09, rgt=f06, tr=f02
        )
        p02 = self.pn(
            c=f02, t=f03, tl=f00, lft=f01, bl=f01, b=f06, br=f10, rgt=f07, tr=f03
        )
        p03 = self.pn(
            c=f03, t=f00, tl=f01, lft=f02, bl=f02, b=f07, br=f11, rgt=f04, tr=f00
        )

        # Assemble the four padded faces on the equator
        p04 = self.pe(
            c=f04,
            t=f00,
            tl=self.tl(f00, f03),
            lft=f03,
            bl=f07,
            b=f11,
            br=self.br(f11, f08),
            rgt=f08,
            tr=f05,
        )
        p05 = self.pe(
            c=f05,
            t=f01,
            tl=self.tl(f01, f00),
            lft=f00,
            bl=f04,
            b=f08,
            br=self.br(f08, f09),
            rgt=f09,
            tr=f06,
        )
        p06 = self.pe(
            c=f06,
            t=f02,
            tl=self.tl(f02, f01),
            lft=f01,
            bl=f05,
            b=f09,
            br=self.br(f09, f10),
            rgt=f10,
            tr=f07,
        )
        p07 = self.pe(
            c=f07,
            t=f03,
            tl=self.tl(f03, f02),
            lft=f02,
            bl=f06,
            b=f10,
            br=self.br(f10, f11),
            rgt=f11,
            tr=f04,
        )

        # Assemble the four padded faces on the southern hemisphere
        p08 = self.ps(
            c=f08, t=f05, tl=f00, lft=f04, bl=f11, b=f11, br=f10, rgt=f09, tr=f09
        )
        p09 = self.ps(
            c=f09, t=f06, tl=f01, lft=f05, bl=f08, b=f08, br=f11, rgt=f10, tr=f10
        )
        p10 = self.ps(
            c=f10, t=f07, tl=f02, lft=f06, bl=f09, b=f09, br=f08, rgt=f11, tr=f11
        )
        p11 = self.ps(
            c=f11, t=f04, tl=f03, lft=f07, bl=f10, b=f10, br=f09, rgt=f08, tr=f08
        )

        res = th.stack(
            (p00, p01, p02, p03, p04, p05, p06, p07, p08, p09, p10, p11), dim=1
        )

        # fold faces into batch dim
        res = self.fold(res)

        try:
            th.cuda.nvtx.range_pop()
        except RuntimeError:
            # Don't bother with NVTX profiling if it's not available
            pass

        return res

    def pn(
        self,
        c: th.Tensor,
        t: th.Tensor,
        tl: th.Tensor,
        lft: th.Tensor,
        bl: th.Tensor,
        b: th.Tensor,
        br: th.Tensor,
        rgt: th.Tensor,
        tr: th.Tensor,
    ) -> th.Tensor:
        """
        Applies padding to a northern hemisphere face c under consideration of its given neighbors.

        Args:
            c: The central face and tensor that is subject for padding
            t: The top neighboring face tensor
            tl: The top left neighboring face tensor
            lft: The left neighboring face tensor
            bl: The bottom left neighboring face tensor
            b: The bottom neighboring face tensor
            br: The bottom right neighboring face tensor
            rgt: The right neighboring face tensor
            tr: The top right neighboring face  tensor

        Returns:
            The padded tensor p
        """
        p = self.p  # Padding size
        d = self.d  # Dimensions for rotations

        # Start with top and bottom to extend the height of the c tensor
        c = th.cat((t.rot90(1, d)[..., -p:, :], c, b[..., :p, :]), dim=-2)

        # Construct the left and right pads including the corner faces
        left = th.cat(
            (
                tl.rot90(2, d)[..., -p:, -p:],
                lft.rot90(-1, d)[..., -p:],
                bl[..., :p, -p:],
            ),
            dim=-2,
        )
        right = th.cat((tr[..., -p:, :p], rgt[..., :p], br[..., :p, :p]), dim=-2)

        return th.cat((left, c, right), dim=-1)

    def pe(
        self,
        c: th.Tensor,
        t: th.Tensor,
        tl: th.Tensor,
        lft: th.Tensor,
        bl: th.Tensor,
        b: th.Tensor,
        br: th.Tensor,
        rgt: th.Tensor,
        tr: th.Tensor,
    ) -> th.Tensor:
        """
        Applies padding to an equatorial face c under consideration of its given neighbors.

        Args:
            c: The central face and tensor that is subject for padding
            t: The top neighboring face tensor
            tl: The top left neighboring face tensor
            lft: The left neighboring face tensor
            bl: The bottom left neighboring face tensor
            b: The bottom neighboring face tensor
            br: The bottom right neighboring face tensor
            rgt: The right neighboring face tensor
            tr: The top right neighboring face  tensor

        Returns
        -------
        th.Tensor:
            The padded tensor p
        """
        p = self.p  # Padding size

        # Start with top and bottom to extend the height of the c tensor
        c = th.cat((t[..., -p:, :], c, b[..., :p, :]), dim=-2)

        # Construct the left and right pads including the corner faces
        left = th.cat((tl[..., -p:, -p:], lft[..., -p:], bl[..., :p, -p:]), dim=-2)
        right = th.cat((tr[..., -p:, :p], rgt[..., :p], br[..., :p, :p]), dim=-2)

        return th.cat((left, c, right), dim=-1)

    def ps(
        self,
        c: th.Tensor,
        t: th.Tensor,
        tl: th.Tensor,
        lft: th.Tensor,
        bl: th.Tensor,
        b: th.Tensor,
        br: th.Tensor,
        rgt: th.Tensor,
        tr: th.Tensor,
    ) -> th.Tensor:
        """
        Applies padding to a southern hemisphere face c under consideration of its given neighbors.

        Args:
            c: The central face and tensor that is subject for padding
            t: The top neighboring face tensor
            tl: The top left neighboring face tensor
            lft: The left neighboring face tensor
            bl: The bottom left neighboring face tensor
            b: The bottom neighboring face tensor
            br: The bottom right neighboring face tensor
            rgt: The right neighboring face tensor
            tr: The top right neighboring face  tensor

        Returns:
            The padded tensor p
        """
        p = self.p  # Padding size
        d = self.d  # Dimensions for rotations

        # Start with top and bottom to extend the height of the c tensor
        c = th.cat((t[..., -p:, :], c, b.rot90(1, d)[..., :p, :]), dim=-2)

        # Construct the left and right pads including the corner faces
        left = th.cat((tl[..., -p:, -p:], lft[..., -p:], bl[..., :p, -p:]), dim=-2)
        right = th.cat(
            (tr[..., -p:, :p], rgt.rot90(-1, d)[..., :p], br.rot90(2, d)[..., :p, :p]),
            dim=-2,
        )

        return th.cat((left, c, right), dim=-1)

    def tl(self, top: th.Tensor, lft: th.Tensor) -> th.Tensor:
        """
        Assembles the top left corner of a center face in the cases where no according top left face is defined on the
        HPX.

        Args:
            top: The face above the center face
            lft: The face left of the center face

        Returns:
            The assembled top left corner (only the sub-part that is required for padding)
        """
        ret = th.zeros_like(top)[..., : self.p, : self.p]  # super ugly but super fast

        # Bottom left point
        ret[..., -1, -1] = 0.5 * top[..., -1, 0] + 0.5 * lft[..., 0, -1]

        # Remaining points
        for i in range(1, self.p):
            ret[..., -i - 1, -i:] = top[
                ..., -i - 1, :i
            ]  # Filling top right above main diagonal
            ret[..., -i:, -i - 1] = lft[
                ..., :i, -i - 1
            ]  # Filling bottom left below main diagonal
            ret[..., -i - 1, -i - 1] = (
                0.5 * top[..., -i - 1, 0] + 0.5 * lft[..., 0, -i - 1]
            )  # Diagonal

        return ret

    def br(self, b: th.Tensor, r: th.Tensor) -> th.Tensor:
        """
        Assembles the bottom right corner of a center face in the cases where no according bottom right face is defined
        on the HPX.

        Args:
            b: The face below the center face
            r: The face right of the center face

        Returns:
            The assembled bottom right corner (only the sub-part that is required for padding)
        """
        ret = th.zeros_like(b)[..., : self.p, : self.p]

        # Top left point
        ret[..., 0, 0] = 0.5 * b[..., 0, -1] + 0.5 * r[..., -1, 0]

        # Remaining points
        for i in range(1, self.p):
            ret[..., :i, i] = r[..., -i:, i]  # Filling top right above main diagonal
            ret[..., i, :i] = b[..., i, -i:]  # Filling bottom left below main diagonal
            ret[..., i, i] = 0.5 * b[..., i, -1] + 0.5 * r[..., -1, i]  # Diagonal

        return ret


class HEALPixLayer(nn.Module):
    """Pytorch module for applying any base torch Module on a HEALPix tensor. Expects all input/output tensors to have a
    shape [..., 12, H, W], where 12 is the dimension of the faces.
    """

    def __init__(self, layer, **kwargs):
        """
        Args:
            layer: Any torch layer function, e.g., nn.Conv2d
            kwargs: The arguments that are passed to the torch layer function, e.g., kernel_size
        """
        super().__init__()
        layers = []

        if "enable_nhwc" in kwargs:
            enable_nhwc = kwargs["enable_nhwc"]
            del kwargs["enable_nhwc"]
        else:
            enable_nhwc = False

        if "enable_healpixpad" in kwargs and kwargs["enable_healpixpad"]:
            raise NotImplementedError(
                "HEALPixPaddingv2 is not available in this environment"
            )

        if "enable_healpixpad" in kwargs:
            del kwargs["enable_healpixpad"]

        if not isinstance(layer, type) or not issubclass(layer, th.nn.Module):
            raise TypeError(
                f"Expected a subclass of torch.nn.Module, got {type(layer).__name__}"
            )
        # Define a HEALPixPadding layer if the given layer is a convolution layer
        if layer.__bases__[0] is nn.modules.conv._ConvNd and kwargs["kernel_size"] > 1:
            kwargs["padding"] = 0  # Disable native padding
            kernel_size = 3 if "kernel_size" not in kwargs else kwargs["kernel_size"]
            dilation = 1 if "dilation" not in kwargs else kwargs["dilation"]
            padding = ((kernel_size - 1) // 2) * dilation
            layers.append(HEALPixPadding(padding=padding, enable_nhwc=enable_nhwc))

        layers.append(layer(**kwargs))
        self.layers = nn.Sequential(*layers)

        if enable_nhwc:
            self.layers = self.layers.to(memory_format=th.channels_last)

    def forward(self, x: th.Tensor) -> th.Tensor:
        """
        Performs the forward pass using the defined layer function and the given data.

        :param x: The input tensor of shape [..., F=12, H, W]
        :return: The output tensor of this HEALPix layer
        """
        res = self.layers(x)
        return res



import dataclasses
from collections import defaultdict
from collections.abc import Mapping
from typing import Literal, Protocol

import numpy as np
import torch
import xarray as xr

from fme.core.dataset.data_typing import VariableMetadata
from fme.core.device import get_device
from fme.core.distributed import Distributed
from fme.core.gridded_ops import GriddedOperations
from fme.core.typing_ import TensorDict, TensorMapping
from fme.core.wandb import Table, WandB


@dataclasses.dataclass
class _SeriesData:
    metric_name: str
    var_name: str
    data: np.ndarray

    def get_wandb_key(self) -> str:
        return f"{self.metric_name}/{self.var_name}"

    def get_xarray_key(self) -> str:
        return f"{self.metric_name}-{self.var_name}"


def get_gen_shape(gen_data: TensorMapping):
    for name in gen_data:
        return gen_data[name].shape


class MeanMetric(Protocol):
    def record(self, target: TensorMapping, gen: TensorMapping, i_time_start: int):
        """
        Update metric for a batch of data.
        """
        ...

    def get(self) -> TensorDict:
        """
        Get the total metric value per variable,
        not divided by number of recorded batches.
        """
        ...


class SingleTargetMeanMetric(Protocol):
    def record(self, tensors: TensorMapping, i_time_start: int):
        """
        Update metric for a batch of data.
        """
        ...

    def get(self) -> TensorDict:
        """
        Get the total metric value per variable,
        not divided by number of recorded batches.
        """
        ...


class AreaWeightedFunction(Protocol):
    """
    A function that computes a metric on the true and predicted values,
    weighted by area.
    """

    def __call__(
        self,
        truth: TensorMapping,
        predicted: TensorMapping,
    ) -> TensorDict: ...


class AreaWeightedSingleTargetFunction(Protocol):
    """
    A function that computes a metric on a single value, weighted by area.
    """

    def __call__(
        self,
        tensors: TensorMapping,
    ) -> TensorDict: ...


def compute_metric_on(
    source: Literal["gen", "target"], metric: AreaWeightedSingleTargetFunction
) -> AreaWeightedFunction:
    """Turns a single-target metric function
    (computed on only the generated or target data) into a function that takes in
    both the generated and target data as arguments, as required for the APIs
    which call generic metric functions.
    """

    def metric_wrapper(
        truth: TensorMapping,
        predicted: TensorMapping,
    ) -> TensorDict:
        if source == "gen":
            return metric(predicted)
        elif source == "target":
            return metric(truth)

    return metric_wrapper


class AreaWeightedReducedMetric:
    """
    A wrapper around an area-weighted metric function.
    """

    def __init__(
        self,
        device: torch.device,
        compute_metric: AreaWeightedFunction,
        n_timesteps: int,
    ):
        self._compute_metric = compute_metric
        self._total: TensorDict = {}
        self._n_batches = torch.zeros(
            n_timesteps, dtype=torch.int32, device=get_device()
        )
        self._device = device
        self._n_timesteps = n_timesteps

    def record(self, target: TensorMapping, gen: TensorMapping, i_time_start: int):
        """Add a batch of data to the metric.

        Args:
            target: Target data. Dictionary mapping variable names to tensors of shape
                [batch, time, height, width].
            gen: Generated data. Dictionary mapping variable names to tensors of shape
                [batch, time, height, width].
            i_time_start: The index of the first timestep in the batch.
        """
        time_dim = 1
        for name in target:
            if target[name].shape != gen[name].shape:
                raise RuntimeError(
                    "Tensors in target and gen must have the same shape, but got "
                    f"{target[name].shape} and {gen[name].shape} "
                    f"for the tensor '{name}'."
                )

        time_dim_len = next(iter(gen.values())).shape[time_dim]
        time_slice = slice(i_time_start, i_time_start + time_dim_len)

        # Update totals for each variable
        new_values = self._compute_metric(truth=target, predicted=gen)
        for name, tensor in new_values.items():
            if name not in self._total:
                self._total[name] = torch.zeros(
                    [self._n_timesteps], dtype=tensor.dtype, device=self._device
                )
            new_value = tensor.mean(dim=0)
            self._total[name][time_slice] += new_value

        self._n_batches[time_slice] += 1

    def get(self) -> TensorDict:
        """Returns the mean metric across recorded batches for each variable."""
        if not self._total:
            # no batches recorded yet
            return defaultdict(lambda: torch.tensor(torch.nan))
        return {name: tensor / self._n_batches for name, tensor in self._total.items()}


class MeanAggregator:
    def __init__(
        self,
        gridded_operations: GriddedOperations,
        target: Literal["norm", "denorm"],
        n_timesteps: int,
        variable_metadata: Mapping[str, VariableMetadata] | None = None,
    ):
        self._gridded_operations = gridded_operations
        # Store one metric object per metric type (e.g., rmse, bias)
        self._target = target
        self._n_timesteps = n_timesteps

        self._dist = Distributed.get_instance()
        if variable_metadata is None:
            self._variable_metadata: Mapping[str, VariableMetadata] = {}
        else:
            self._variable_metadata = variable_metadata

        self._variable_metrics: dict[str, MeanMetric] = {}
        device = get_device()

        self._variable_metrics["weighted_rmse"] = AreaWeightedReducedMetric(
            device=device,
            compute_metric=self._gridded_operations.area_weighted_rmse_dict,
            n_timesteps=self._n_timesteps,
        )
        if self._target == "denorm":
            self._variable_metrics["weighted_grad_mag_percent_diff"] = (
                AreaWeightedReducedMetric(
                    device=device,
                    compute_metric=self._gridded_operations.area_weighted_gradient_magnitude_percent_diff_dict,  # noqa: E501
                    n_timesteps=self._n_timesteps,
                )
            )
        self._variable_metrics["weighted_mean_gen"] = AreaWeightedReducedMetric(
            device=device,
            compute_metric=compute_metric_on(
                source="gen",
                metric=(
                    lambda tensors: self._gridded_operations.area_weighted_mean_dict(
                        tensors
                    )
                ),
            ),
            n_timesteps=self._n_timesteps,
        )
        self._variable_metrics["weighted_mean_target"] = AreaWeightedReducedMetric(
            device=device,
            compute_metric=compute_metric_on(
                source="target",
                metric=(
                    lambda tensors: self._gridded_operations.area_weighted_mean_dict(
                        tensors
                    )
                ),
            ),
            n_timesteps=self._n_timesteps,
        )
        self._variable_metrics["weighted_bias"] = AreaWeightedReducedMetric(
            device=device,
            compute_metric=self._gridded_operations.area_weighted_mean_bias_dict,
            n_timesteps=self._n_timesteps,
        )
        self._variable_metrics["weighted_std_gen"] = AreaWeightedReducedMetric(
            device=device,
            compute_metric=compute_metric_on(
                source="gen",
                metric=(
                    lambda tensors: self._gridded_operations.area_weighted_std_dict(
                        tensors
                    )
                ),
            ),
            n_timesteps=self._n_timesteps,
        )
        self._n_batches = 0

    @torch.no_grad()
    def record_batch(
        self,
        target_data: TensorMapping,
        gen_data: TensorMapping,
        target_data_norm: TensorMapping,
        gen_data_norm: TensorMapping,
        i_time_start: int = 0,
    ):
        if self._target == "norm":
            target_data = target_data_norm
            gen_data = gen_data_norm
        for metric in self._variable_metrics.values():
            metric.record(
                target=target_data,
                gen=gen_data,
                i_time_start=i_time_start,
            )
        self._n_batches += 1

    def _get_series_data(self, step_slice: slice | None = None) -> list[_SeriesData]:
        """Converts internally stored variable_metrics to a list."""
        if self._n_batches == 0:
            raise ValueError("No batches have been recorded.")
        data: list[_SeriesData] = []
        for name, metric in self._variable_metrics.items():
            metric_results = metric.get()  # TensorDict: {var_name: metric_series}
            sorted_keys = sorted(list(metric_results.keys()))
            for key in sorted_keys:
                arr = metric_results[key].detach()
                if step_slice is not None:
                    arr = arr[step_slice]
                datum = _SeriesData(
                    metric_name=name,
                    var_name=key,
                    data=self._dist.reduce_mean(arr).cpu().numpy(),
                )
                data.append(datum)
        return data

    @torch.no_grad()
    def get_logs(self, label: str, step_slice: slice | None = None):
        """
        Returns logs as can be reported to WandB.

        Args:
            label: Label to prepend to all log keys.
            step_slice: Slice of forecast steps to log.
        """
        logs = {}
        series_data: dict[str, np.ndarray] = {
            datum.get_wandb_key(): datum.data
            for datum in self._get_series_data(step_slice)
        }
        init_step = 0 if step_slice is None else step_slice.start
        table = data_to_table(series_data, init_step)
        logs[f"{label}/series"] = table
        return logs

    @torch.no_grad()
    def get_dataset(self) -> xr.Dataset:
        """
        Returns a dataset representation of the logs.
        """
        data_vars = {}
        for datum in self._get_series_data():
            metadata = self._variable_metadata.get(
                datum.var_name, VariableMetadata("unknown_units", datum.var_name)
            )
            data_vars[datum.get_xarray_key()] = xr.DataArray(
                datum.data, dims=["forecast_step"], attrs=metadata._asdict()
            )

        if len(data_vars.values()) > 0:
            n_forecast_steps = len(next(iter(data_vars.values())))
            coords = {"forecast_step": np.arange(n_forecast_steps)}
        else:
            coords = {"forecast_step": np.arange(0)}

        return xr.Dataset(data_vars=data_vars, coords=coords)


def data_to_table(data: dict[str, np.ndarray], init_step: int = 0) -> Table:
    """
    Convert a dictionary of 1-dimensional timeseries data to a wandb Table.

    Args:
        data: dictionary of timeseries data.
        init_step: initial step corresponding to the first row's "forecast_step"
    """
    keys = sorted(list(data.keys()))
    wandb = WandB.get_instance()
    table = wandb.Table(columns=["forecast_step"] + keys)
    if len(keys) > 0:
        for i in range(len(data[keys[0]])):
            row = [init_step + i]
            for key in keys:
                row.append(data[key][i])
            table.add_data(*row)
    return table


class AreaWeightedSingleTargetReducedMetric:
    """
    A wrapper around an area-weighted metric function on a single data source.
    """

    def __init__(
        self,
        device: torch.device,
        compute_metric: AreaWeightedSingleTargetFunction,
        n_timesteps: int,
    ):
        self._compute_metric = compute_metric
        self._total: TensorDict = {}
        self._n_batches = torch.zeros(
            n_timesteps, dtype=torch.int32, device=get_device()
        )
        self._device = device
        self._n_timesteps = n_timesteps

    def record(self, tensors: TensorMapping, i_time_start: int):
        """Add a batch of data to the metric.

        Args:
            tensors: Dictionary mapping variable names to tensors of shape
                [batch, time, height, width].
            i_time_start: The index of the first timestep in the batch.
        """
        time_dim = 1

        time_dim_len = next(iter(tensors.values())).shape[time_dim]
        time_slice = slice(i_time_start, i_time_start + time_dim_len)

        # Update totals for each variable
        new_values = self._compute_metric(tensors)
        for name, tensor in new_values.items():
            if name not in self._total:
                self._total[name] = torch.zeros(
                    [self._n_timesteps], dtype=tensor.dtype, device=self._device
                )
            new_value = tensor.mean(dim=0)
            self._total[name][time_slice] += new_value

        self._n_batches[time_slice] += 1

    def get(self) -> TensorDict:
        """Returns the mean metric across recorded batches for each variable."""
        if not self._total:
            return defaultdict(lambda: torch.tensor(torch.nan))
        return {name: tensor / self._n_batches for name, tensor in self._total.items()}


class SingleTargetMeanAggregator:
    def __init__(
        self,
        gridded_operations: GriddedOperations,
        n_timesteps: int,
        variable_metadata: Mapping[str, VariableMetadata] | None = None,
    ):
        self._ops = gridded_operations
        self._n_timesteps = n_timesteps

        self._dist = Distributed.get_instance()
        if variable_metadata is None:
            self._variable_metadata: Mapping[str, VariableMetadata] = {}
        else:
            self._variable_metadata = variable_metadata

        self._variable_metrics: dict[str, SingleTargetMeanMetric] = {}
        device = get_device()

        self._variable_metrics["weighted_mean_gen"] = (
            AreaWeightedSingleTargetReducedMetric(
                device=device,
                compute_metric=(
                    lambda tensors: self._ops.area_weighted_mean_dict(tensors)
                ),
                n_timesteps=self._n_timesteps,
            )
        )
        self._variable_metrics["weighted_std_gen"] = (
            AreaWeightedSingleTargetReducedMetric(
                device=device,
                compute_metric=(
                    lambda tensors: self._ops.area_weighted_std_dict(tensors)
                ),
                n_timesteps=self._n_timesteps,
            )
        )
        self._n_batches = 0

    @torch.no_grad()
    def record_batch(
        self,
        data: TensorMapping,
        i_time_start: int = 0,
    ):
        for metric in self._variable_metrics.values():
            metric.record(
                tensors=data,
                i_time_start=i_time_start,
            )
        self._n_batches += 1

    def _get_series_data(self, step_slice: slice | None = None) -> list[_SeriesData]:
        """Converts internally stored variable_metrics to a list."""
        if self._n_batches == 0:
            raise ValueError("No batches have been recorded.")
        data: list[_SeriesData] = []
        for name, metric in self._variable_metrics.items():
            metric_results = metric.get()  # TensorDict: {var_name: metric_series}
            sorted_keys = sorted(list(metric_results.keys()))
            for key in sorted_keys:
                arr = metric_results[key].detach()
                if step_slice is not None:
                    arr = arr[step_slice]
                datum = _SeriesData(
                    metric_name=name,
                    var_name=key,
                    data=self._dist.reduce_mean(arr).cpu().numpy(),
                )
                data.append(datum)
        return data

    @torch.no_grad()
    def get_logs(self, label: str, step_slice: slice | None = None):
        """
        Returns logs as can be reported to WandB.

        Args:
            label: Label to prepend to all log keys.
            step_slice: Slice of forecast steps to log.
        """
        logs = {}
        series_data: dict[str, np.ndarray] = {
            datum.get_wandb_key(): datum.data
            for datum in self._get_series_data(step_slice)
        }
        init_step = 0 if step_slice is None else step_slice.start
        table = data_to_table(series_data, init_step)
        logs[f"{label}/series"] = table
        return logs

    @torch.no_grad()
    def get_dataset(self) -> xr.Dataset:
        """
        Returns a dataset representation of the logs.
        """
        data_vars = {}
        for datum in self._get_series_data():
            metadata = self._variable_metadata.get(
                datum.var_name, VariableMetadata("unknown_units", datum.var_name)
            )
            data_vars[datum.get_xarray_key()] = xr.DataArray(
                datum.data, dims=["forecast_step"], attrs=metadata._asdict()
            )

        n_forecast_steps = len(next(iter(data_vars.values())))
        coords = {"forecast_step": np.arange(n_forecast_steps)}
        return xr.Dataset(data_vars=data_vars, coords=coords)



import torch

from fme.ace.aggregator.inference.zonal_mean import ZonalMeanAggregator
from fme.core import get_device

n_sample, n_time, ny, nx = 3, 6, 10, 20


def zonal_mean(data: torch.Tensor) -> torch.Tensor:
    return data.mean(dim=3)


def test_zonal_mean_dims():
    agg = ZonalMeanAggregator(zonal_mean, n_timesteps=n_time)
    target_data = {"a": torch.randn(n_sample, n_time, ny, nx, device=get_device())}
    gen_data = {"a": torch.randn(n_sample, n_time, ny, nx, device=get_device())}
    agg.record_batch(target_data, gen_data, target_data, gen_data, i_time_start=0)
    for data in (agg._target_data, agg._gen_data):
        assert data is not None
        assert data["a"].size() == (
            n_sample,
            n_time,
            ny,
        )


def test_zonal_mean_lat_varying():
    agg = ZonalMeanAggregator(zonal_mean, n_timesteps=n_time)
    arr = torch.arange(ny, dtype=torch.float32, device=get_device())
    arr = arr[None, None, :, None].expand(n_sample, n_time, -1, nx)
    agg.record_batch({"a": arr}, {"a": arr}, {"a": arr}, {"a": arr}, i_time_start=0)
    for data in (agg._target_data, agg._gen_data):
        assert data is not None
        torch.testing.assert_close(
            data["a"][0, 0, :],  # one time row of the zonal mean
            torch.arange(ny, dtype=torch.float32, device=get_device()),
        )


def test_zonal_mean_zonally_varying():
    agg = ZonalMeanAggregator(zonal_mean, n_timesteps=n_time)
    arr = torch.arange(nx, dtype=torch.float32, device=get_device())
    arr = arr[None, None, None, :].expand(n_sample, n_time, ny, -1)
    agg.record_batch({"a": arr}, {"a": arr}, {"a": arr}, {"a": arr}, i_time_start=0)
    for data in (agg._target_data, agg._gen_data):
        assert data is not None
        torch.testing.assert_close(
            data["a"][0, 0, :],  # one time row of the zonal mean
            arr.mean() * torch.ones(ny, dtype=torch.float32, device=get_device()),
        )


def test_zonal_mean_batch_varying():
    agg = ZonalMeanAggregator(zonal_mean, n_timesteps=n_time)
    for i in range(n_sample):  # assume one sample per batch
        arr = torch.tensor(i, dtype=torch.float32, device=get_device())
        arr = arr[None, None, None, None].expand(-1, n_time, ny, nx)
        agg.record_batch({"a": arr}, {"a": arr}, {"a": arr}, {"a": arr}, i_time_start=0)
    for data in (agg._target_data, agg._gen_data):
        assert data is not None
        torch.testing.assert_close(
            data["a"].sum(dim=0)[0, 0],  # sum over batches, then pick a time/lat point
            torch.arange(n_sample, dtype=torch.float32, device=get_device()).sum(),
            # should be same as sum over batches
        )


def test_zonal_mean_mulitple_time_slices():
    n_time_windows = 2
    n_time_in_memory = n_time // n_time_windows
    agg = ZonalMeanAggregator(zonal_mean, n_timesteps=n_time)
    for i_time in range(0, n_time, n_time_in_memory):
        arr = torch.arange(ny, dtype=torch.float32, device=get_device())
        arr = arr[None, None, :, None].expand(n_sample, n_time_in_memory, ny, nx)
        agg.record_batch(
            {"a": arr}, {"a": arr}, {"a": arr}, {"a": arr}, i_time_start=i_time
        )
    for data in (agg._target_data, agg._gen_data):
        assert data is not None
        torch.testing.assert_close(
            (data["a"] / agg._n_batches)[0, 0, :],
            torch.arange(ny, dtype=torch.float32, device=get_device()),
        )



"""This file contains unit tests related to creating torch Datasets from climate
data (e.g. netCDF files)."""

import math
import os
import pathlib

import cftime
import numpy as np
import pytest
import torch
import xarray as xr

import fme
from fme.ace.data_loading.batch_data import BatchData, PrognosticState
from fme.ace.data_loading.config import (
    ConcatDatasetConfig,
    DataLoaderConfig,
    MergeDatasetConfig,
)
from fme.ace.data_loading.getters import (
    get_forcing_data,
    get_gridded_data,
    get_inference_data,
)
from fme.ace.data_loading.inference import (
    ExplicitIndices,
    ForcingDataLoaderConfig,
    InferenceDataLoaderConfig,
    InferenceDataset,
    InferenceInitialConditionIndices,
    TimestampList,
)
from fme.ace.data_loading.perturbation import PerturbationSelector, SSTPerturbation
from fme.ace.requirements import DataRequirements, PrognosticStateDataRequirements
from fme.core.coordinates import HybridSigmaPressureCoordinate
from fme.core.dataset.config import MergeNoConcatDatasetConfig, XarrayDataConfig
from fme.core.typing_ import Slice


def _get_coords(dim_sizes, calendar, timestep_size=1, timestep_start=0):
    coords = {}
    for dim_name, size in dim_sizes.items():
        if dim_name == "time":
            dtype = np.int64
            start = timestep_start
            step = timestep_size
            size = size * step
            attrs = {"calendar": calendar, "units": "days since 1970-01-01"}
        else:
            dtype = np.float32
            start = 0
            step = 1
            attrs = {}
        coord_value = np.arange(start, start + size, step, dtype=dtype)
        coord = xr.DataArray(coord_value, dims=(dim_name,), attrs=attrs)
        coords[dim_name] = coord
    return coords


def _save_netcdf(
    filename, dim_sizes, variable_names, calendar, timestep_size=1, timestep_start=0
):
    data_vars = {}
    for name in variable_names:
        if name == "constant_mask":
            data = np.ones(list(dim_sizes.values()))
        else:
            data = np.random.randn(*list(dim_sizes.values()))
        if len(dim_sizes) > 0:
            data = data.astype(np.float32)  # type: ignore
        data_vars[name] = xr.DataArray(
            data, dims=list(dim_sizes), attrs={"units": "m", "long_name": name}
        )
    coords = _get_coords(dim_sizes, calendar, timestep_size, timestep_start)
    for i in range(7):
        data_vars[f"ak_{i}"] = float(i)
        data_vars[f"bk_{i}"] = float(i + 1)
    ds = xr.Dataset(data_vars=data_vars, coords=coords)
    ds.to_netcdf(filename, unlimited_dims=["time"], format="NETCDF4_CLASSIC")
    return ds


def _create_dataset_on_disk(
    data_dir: pathlib.Path,
    calendar: str = "proleptic_gregorian",
    data_dim_sizes=None,
    n_times: int = 3,
    timestep_size: int = 1,
    in_variable_names=["foo", "bar", "baz"],
    out_variable_names=["foo", "bar"],
    filename="data.nc",
) -> pathlib.Path:
    if data_dim_sizes is None:
        data_dim_sizes = {"time": n_times, "lat": 16, "lon": 32}
    seed = 0
    np.random.seed(seed)
    mask_name = "mask"
    constant_mask_name = "constant_mask"
    all_variable_names = list(set(in_variable_names + out_variable_names)) + [
        mask_name,
        constant_mask_name,
    ]

    data_path = data_dir / filename
    _save_netcdf(data_path, data_dim_sizes, all_variable_names, calendar, timestep_size)

    return data_path


def test_ensemble_loader(tmp_path, num_ensemble_members=3):
    """Tests that the ensemble loader returns the correct number of samples."""

    # Create a dataset for each ensemble member. We assume that each member
    # corresponds to an initial condition.
    netcdfs: list[pathlib.Path] = []
    for i in range(num_ensemble_members):
        ic_path = tmp_path / f"ic{i}"
        ic_path.mkdir()
        _create_dataset_on_disk(ic_path)
        netcdfs.append(ic_path)

    config = DataLoaderConfig(
        dataset=ConcatDatasetConfig(
            concat=[XarrayDataConfig(data_path=str(path)) for path in netcdfs]
        ),
        batch_size=1,
        num_data_workers=0,
    )
    window_timesteps = 2  # 1 initial condition and 1 step forward
    requirements = DataRequirements(["foo"], window_timesteps)

    n_timesteps = 3  # hard coded to match `_create_dataset_on_disk`.
    samples_per_member = n_timesteps - window_timesteps + 1

    data = get_gridded_data(config, True, requirements)
    assert data.n_batches == samples_per_member * num_ensemble_members
    assert isinstance(data._vertical_coordinate, HybridSigmaPressureCoordinate)


def test_ensemble_loader_n_samples(tmp_path, num_ensemble_members=3, n_samples=1):
    """Tests that the ensemble loader returns the correct number of samples
    when n_samples is set in config.
    """

    # Create a dataset for each ensemble member. We assume that each member
    # corresponds to an initial condition.
    netcdfs: list[pathlib.Path] = []
    for i in range(num_ensemble_members):
        ic_path = tmp_path / f"ic{i}"
        ic_path.mkdir()
        _create_dataset_on_disk(ic_path)
        netcdfs.append(ic_path)

    config = DataLoaderConfig(
        dataset=ConcatDatasetConfig(
            concat=[
                XarrayDataConfig(data_path=str(path), subset=Slice(stop=n_samples))
                for path in netcdfs
            ]
        ),
        batch_size=1,
        num_data_workers=0,
    )

    window_timesteps = 2  # 1 initial condition and 1 step forward
    requirements = DataRequirements(["foo"], window_timesteps)

    data = get_gridded_data(config, True, requirements)
    assert data.n_batches == n_samples * num_ensemble_members
    assert isinstance(data._vertical_coordinate, HybridSigmaPressureCoordinate)


def test_xarray_loader(tmp_path):
    """Checks that vertical coordinates are present."""
    _create_dataset_on_disk(tmp_path)
    config = DataLoaderConfig(
        dataset=ConcatDatasetConfig(
            concat=[XarrayDataConfig(data_path=tmp_path, n_repeats=1)]
        ),
        batch_size=1,
        num_data_workers=0,
    )
    window_timesteps = 2  # 1 initial condition and 1 step forward
    requirements = DataRequirements(["foo"], window_timesteps)
    data = get_gridded_data(config, True, requirements)  # type: ignore
    assert isinstance(data._vertical_coordinate, HybridSigmaPressureCoordinate)
    assert data._vertical_coordinate.ak.device == fme.get_device()


def test_xarray_loader_sample_with_replacement(tmp_path):
    _create_dataset_on_disk(tmp_path, n_times=3)
    config = DataLoaderConfig(
        dataset=ConcatDatasetConfig(
            concat=[XarrayDataConfig(data_path=tmp_path, n_repeats=1)]
        ),
        batch_size=1,
        num_data_workers=0,
        sample_with_replacement=10,
    )
    window_timesteps = 2  # 1 initial condition and 1 step forward
    requirements = DataRequirements(["foo"], window_timesteps)
    data = get_gridded_data(config, True, requirements)  # type: ignore
    assert isinstance(data._vertical_coordinate, HybridSigmaPressureCoordinate)
    assert data._vertical_coordinate.ak.device == fme.get_device()
    epoch_samples = list(data.loader)
    assert len(epoch_samples) == 10


def test_xarray_loader_using_merged_dataset(tmp_path, tmp_path_factory):
    _create_dataset_on_disk(tmp_path)
    other_path = tmp_path_factory.mktemp("other")
    _create_dataset_on_disk(
        other_path,
        in_variable_names=["foo2", "bar2"],
        out_variable_names=["foo2", "bar2"],
        filename="other_source.nc",
    )

    config = DataLoaderConfig(
        dataset=MergeDatasetConfig(
            merge=[
                XarrayDataConfig(
                    data_path=tmp_path,
                    file_pattern="data*.nc",
                    n_repeats=1,
                ),
                XarrayDataConfig(
                    data_path=other_path, file_pattern="other_source*.nc", n_repeats=1
                ),
            ]
        ),
        batch_size=1,
        num_data_workers=0,
    )
    window_timesteps = 2  # 1 initial condition and 1 step forward
    requirements = DataRequirements(["foo", "foo2"], window_timesteps)
    data = get_gridded_data(config, True, requirements)  # type: ignore
    assert "foo" in data.variable_metadata.keys()
    assert "foo2" in data.variable_metadata.keys()


def test_xarray_loader_using_merged_dataset_errors_if_different_time(
    tmp_path, tmp_path_factory
):
    _create_dataset_on_disk(tmp_path)
    other_path = tmp_path_factory.mktemp("other")
    _create_dataset_on_disk(
        other_path,
        in_variable_names=["foo2", "bar2"],
        out_variable_names=["foo2", "bar2"],
        filename="other_source.nc",
        n_times=4,
    )

    config = DataLoaderConfig(
        dataset=MergeDatasetConfig(
            merge=[
                XarrayDataConfig(
                    data_path=tmp_path,
                    file_pattern="data*.nc",
                    n_repeats=1,
                ),
                XarrayDataConfig(
                    data_path=other_path,
                    file_pattern="other_source*.nc",
                    n_repeats=1,
                ),
            ]
        ),
        batch_size=1,
        num_data_workers=0,
    )
    window_timesteps = 2  # 1 initial condition and 1 step forward
    requirements = DataRequirements(["foo", "foo2"], window_timesteps)
    with pytest.raises(
        ValueError,
        match="All datasets in a merged dataset must have the same sample start times",
    ):
        get_gridded_data(config, True, requirements)  # type: ignore

    # subset source2 to have the same time stamps as source1
    config = DataLoaderConfig(
        dataset=MergeDatasetConfig(
            merge=[
                XarrayDataConfig(
                    data_path=tmp_path,
                    file_pattern="data*.nc",
                    n_repeats=1,
                ),
                XarrayDataConfig(
                    data_path=other_path,
                    file_pattern="other_source*.nc",
                    n_repeats=1,
                    subset=Slice(stop=2),
                ),
            ]
        ),
        batch_size=1,
        num_data_workers=0,
    )
    get_gridded_data(config, True, requirements)  # type: ignore


def test_xarray_loader_hpx(tmp_path):
    """Checks that vertical coordinates are present."""
    n_times = 3
    data_dim_sizes = {"time": n_times, "face": 12, "width": 16, "height": 16}
    _create_dataset_on_disk(tmp_path, data_dim_sizes=data_dim_sizes, n_times=n_times)
    config = DataLoaderConfig(
        dataset=ConcatDatasetConfig(
            concat=[
                XarrayDataConfig(
                    data_path=tmp_path, n_repeats=1, spatial_dimensions="healpix"
                )
            ]
        ),
        batch_size=1,
        num_data_workers=0,
    )
    window_timesteps = 2  # 1 initial condition and 1 step forward
    requirements = DataRequirements(["foo"], window_timesteps)
    data = get_gridded_data(config, True, requirements)  # type: ignore
    for batch in data.loader:
        assert batch is not None
        # expect healpix shape
        assert batch.data["foo"].shape == (1, window_timesteps, 12, 16, 16)
        break
    assert isinstance(data._vertical_coordinate, HybridSigmaPressureCoordinate)
    assert data._vertical_coordinate.ak.device == fme.get_device()


def test_loader_n_repeats_but_not_infer_timestep_error(tmp_path):
    _create_dataset_on_disk(tmp_path)
    with pytest.raises(ValueError, match="infer_timestep must be True"):
        DataLoaderConfig(
            dataset=ConcatDatasetConfig(
                concat=[
                    XarrayDataConfig(
                        data_path=tmp_path, n_repeats=2, infer_timestep=False
                    )
                ]
            ),
            batch_size=1,
            num_data_workers=0,
        )


def test_inference_data_loader(tmp_path):
    _create_dataset_on_disk(tmp_path, n_times=14)
    batch_size = 2
    step = 7
    config = InferenceDataLoaderConfig(
        dataset=XarrayDataConfig(
            data_path=tmp_path,
            n_repeats=1,
        ),
        start_indices=InferenceInitialConditionIndices(
            first=0, n_initial_conditions=batch_size, interval=step
        ),
    )
    n_forward_steps_in_memory = 3
    window_requirements = DataRequirements(
        names=["foo", "bar"],
        n_timesteps=n_forward_steps_in_memory + 1,
    )
    initial_condition_requirements = PrognosticStateDataRequirements(
        names=["foo"],
        n_timesteps=1,
    )
    data = get_inference_data(
        config,
        total_forward_steps=6,
        window_requirements=window_requirements,
        initial_condition=initial_condition_requirements,
    )
    data_loader = data.loader
    batch_data = next(iter(data_loader))
    assert isinstance(batch_data, BatchData)
    for name in ["foo", "bar"]:
        assert isinstance(batch_data.data[name], torch.Tensor)
        assert batch_data.data[name].shape == (
            batch_size,
            n_forward_steps_in_memory + 1,
            16,
            32,
        )
    assert isinstance(batch_data.time, xr.DataArray)
    assert list(batch_data.time.dims) == ["sample", "time"]
    assert batch_data.time.sizes["sample"] == batch_size
    assert batch_data.time.sizes["time"] == n_forward_steps_in_memory + 1
    assert batch_data.time.dt.calendar == "proleptic_gregorian"
    assert isinstance(data._vertical_coordinate, HybridSigmaPressureCoordinate)
    assert data._vertical_coordinate.ak.device == fme.get_device()
    initial_condition = data.initial_condition.as_batch_data()
    assert isinstance(initial_condition, BatchData)
    assert "bar" not in initial_condition.data
    assert initial_condition.data["foo"].shape == (batch_size, 1, 16, 32)


@pytest.fixture(params=["julian", "proleptic_gregorian", "noleap"])
def calendar(request):
    """
    These are the calendars for the datasets we tend to use: 'julian'
    for FV3GFS, 'noleap' for E3SM, and 'proleptic_gregorian' for generic
    datetimes in testing.

    Check that datasets created with each calendar for their time coordinate
    are read by the data loader and the calendar is retained.
    """
    return request.param


def test_data_loader_outputs(tmp_path, calendar):
    _create_dataset_on_disk(tmp_path, calendar=calendar)
    n_samples = 2
    config = DataLoaderConfig(
        dataset=ConcatDatasetConfig(
            concat=[XarrayDataConfig(data_path=tmp_path, subset=Slice(stop=n_samples))]
        ),
        batch_size=n_samples,
        num_data_workers=0,
    )
    window_timesteps = 2  # 1 initial condition and 1 step forward
    requirements = DataRequirements(["foo"], window_timesteps)
    data = get_gridded_data(config, True, requirements)  # type: ignore
    batch_data = next(iter(data.loader))
    assert isinstance(batch_data, BatchData)
    assert isinstance(batch_data.data["foo"], torch.Tensor)
    assert batch_data.data["foo"].shape[0] == n_samples
    assert isinstance(batch_data.time, xr.DataArray)
    assert list(batch_data.time.dims) == ["sample", "time"]
    assert batch_data.time.sizes["sample"] == n_samples
    assert batch_data.time.sizes["time"] == window_timesteps
    assert batch_data.time.dt.calendar == calendar


@pytest.mark.parametrize(
    ("first_ic_index,n_initial_conditions,ic_interval,num_forward_steps,raises_error"),
    [
        (0, 1, 1, 9, False),
        (0, 1, 1, 10, True),
        (1, 1, 1, 8, False),
        (1, 1, 1, 9, True),
        (2, 3, 2, 3, False),
        (2, 3, 2, 4, True),
    ],
)
def test_inference_data_loader_validate_n_forward_steps(
    tmp_path,
    first_ic_index,
    n_initial_conditions,
    ic_interval,
    num_forward_steps,
    raises_error,
):
    """Check exception is raised if n_forward_steps exceeds number of
    forward steps in dataset available after last initial condition."""

    total_dataset_timesteps = 10
    _create_dataset_on_disk(tmp_path, n_times=total_dataset_timesteps)
    start_indices = InferenceInitialConditionIndices(
        first=first_ic_index,
        n_initial_conditions=n_initial_conditions,
        interval=ic_interval,
    )
    config = InferenceDataLoaderConfig(
        dataset=XarrayDataConfig(
            data_path=tmp_path,
            n_repeats=1,
        ),
        start_indices=start_indices,
    )
    n_forward_steps_in_memory = num_forward_steps
    window_requirements = DataRequirements(
        names=["foo", "bar"],
        n_timesteps=n_forward_steps_in_memory + 1,
    )
    initial_condition_requirements = PrognosticStateDataRequirements(
        names=["foo"],
        n_timesteps=1,
    )

    if raises_error:
        with pytest.raises(ValueError):
            get_inference_data(
                config,
                total_forward_steps=num_forward_steps,
                window_requirements=window_requirements,
                initial_condition=initial_condition_requirements,
            )
    else:
        get_inference_data(
            config,
            total_forward_steps=num_forward_steps,
            window_requirements=window_requirements,
            initial_condition=initial_condition_requirements,
        )


@pytest.mark.parametrize(
    "start, stop, batch_size, raises_error",
    [
        pytest.param(0, 3, 1, False, id="valid"),
        pytest.param(10000, 10100, 1, True, id="no_samples"),
        pytest.param(0, 25, 50, True, id="batch_size_larger_than_nsamples"),
    ],
)
def test_zero_batches_raises_error(tmp_path, start, stop, batch_size, raises_error):
    _create_dataset_on_disk(tmp_path)
    config = DataLoaderConfig(
        dataset=ConcatDatasetConfig(
            concat=[
                XarrayDataConfig(
                    data_path=tmp_path, n_repeats=10, subset=Slice(start, stop)
                )
            ]
        ),
        batch_size=batch_size,
        num_data_workers=0,
    )
    window_timesteps = 2  # 1 initial condition and 1 step forward
    requirements = DataRequirements(["foo"], window_timesteps)
    if raises_error:
        with pytest.raises(ValueError):
            get_gridded_data(config, True, requirements)  # type: ignore
    else:
        get_gridded_data(config, True, requirements)  # type: ignore


@pytest.mark.parametrize("n_initial_conditions", [1, 2])
def test_get_forcing_data(tmp_path, n_initial_conditions):
    calendar = "proleptic_gregorian"
    total_forward_steps = 5
    forward_steps_in_memory = 2
    _create_dataset_on_disk(tmp_path, calendar=calendar, n_times=10)
    config = ForcingDataLoaderConfig(dataset=XarrayDataConfig(data_path=tmp_path))
    window_requirements = DataRequirements(
        names=["foo"],
        n_timesteps=forward_steps_in_memory + 1,
    )
    time_values = [
        [cftime.datetime(1970, 1, 1 + 2 * n, calendar=calendar)]
        for n in range(n_initial_conditions)
    ]
    initial_condition = BatchData.new_on_cpu(
        data={"foo": torch.randn(n_initial_conditions, 1, 1, 1)},
        time=xr.DataArray(time_values, dims=["sample", "time"]),
    )
    data = get_forcing_data(
        config,
        total_forward_steps,
        window_requirements=window_requirements,
        initial_condition=PrognosticState(initial_condition),
    )
    assert len(data._loader.dataset) == math.ceil(  # type: ignore
        total_forward_steps / forward_steps_in_memory
    )
    batch_data = next(iter(data.loader))
    assert isinstance(batch_data, BatchData)
    assert isinstance(batch_data.data["foo"], torch.Tensor)
    assert set(batch_data.data.keys()) == {"foo"}
    assert batch_data.data["foo"].shape[0] == len(time_values)
    assert batch_data.data["foo"].shape[1] == forward_steps_in_memory + 1
    assert list(batch_data.time.dims) == ["sample", "time"]
    xr.testing.assert_allclose(batch_data.time[:, 0], initial_condition.time[:, 0])
    assert batch_data.time.dt.calendar == calendar
    xr.testing.assert_equal(
        data.initial_condition.as_batch_data().time,
        initial_condition.time,
    )
    np.testing.assert_allclose(
        data.initial_condition.as_batch_data().data["foo"].cpu().numpy(),
        initial_condition.data["foo"].cpu().numpy(),
    )


def test_inference_loader_raises_if_subset():
    with pytest.raises(ValueError):
        InferenceDataLoaderConfig(
            dataset=XarrayDataConfig(data_path="foo", subset=Slice(stop=1)),
            start_indices=ExplicitIndices([0, 1]),
        )


def test_inference_loader_merged_inputs(tmp_path, tmp_path_factory):
    _create_dataset_on_disk(tmp_path)
    other_path = tmp_path_factory.mktemp("other")
    _create_dataset_on_disk(
        other_path,
        in_variable_names=["foo2", "bar2"],
        out_variable_names=["foo2", "bar2"],
        filename="other_source.nc",
    )
    merged = InferenceDataLoaderConfig(
        dataset=MergeNoConcatDatasetConfig(
            merge=[
                XarrayDataConfig(data_path=tmp_path),
                XarrayDataConfig(data_path=other_path),
            ]
        ),
        start_indices=ExplicitIndices([0, 1]),
    )
    window_requirements = DataRequirements(
        names=["foo", "foo2"],
        n_timesteps=2,
    )
    initial_condition_requirements = PrognosticStateDataRequirements(
        names=["foo", "foo2"],
        n_timesteps=1,
    )
    data = get_inference_data(
        merged,
        total_forward_steps=1,
        window_requirements=window_requirements,
        initial_condition=initial_condition_requirements,
        surface_temperature_name="foo",
        ocean_fraction_name="constant_mask",
    )
    data_loader = data.loader
    batch_data = next(iter(data_loader))
    assert "foo" in batch_data.data.keys()
    assert "foo2" in batch_data.data.keys()


def test_forcing_loader_raises_if_subset():
    with pytest.raises(ValueError):
        ForcingDataLoaderConfig(
            dataset=XarrayDataConfig(data_path="foo", subset=Slice(stop=1))
        )


def test_forcing_loader_config_merged_dataset_inputs():
    ForcingDataLoaderConfig(
        dataset=MergeNoConcatDatasetConfig(
            merge=[
                XarrayDataConfig(data_path="foo"),
                XarrayDataConfig(data_path="bar"),
            ]
        )
    )


def test_forcing_loader_loads_merged_dataset(tmp_path, tmp_path_factory):
    calendar = "proleptic_gregorian"
    _create_dataset_on_disk(tmp_path, calendar=calendar)
    other_path = tmp_path_factory.mktemp("other")
    _create_dataset_on_disk(
        other_path,
        calendar=calendar,
        in_variable_names=["foo2", "bar2"],
        out_variable_names=["foo2", "bar2"],
    )
    config = ForcingDataLoaderConfig(dataset=XarrayDataConfig(data_path=tmp_path))
    config = ForcingDataLoaderConfig(
        dataset=MergeNoConcatDatasetConfig(
            merge=[
                XarrayDataConfig(data_path=tmp_path),
                XarrayDataConfig(data_path=other_path),
            ]
        ),
    )
    window_requirements = DataRequirements(
        names=["foo", "foo2"],
        n_timesteps=2,
    )
    time_values = [[cftime.datetime(1970, 1, 1, calendar=calendar)]]
    initial_condition = BatchData.new_on_cpu(
        data={"foo": torch.randn(1, 1, 1, 1), "foo2": torch.randn(1, 1, 1, 1)},
        time=xr.DataArray(time_values, dims=["sample", "time"]),
    )
    data = get_forcing_data(
        config,
        1,
        window_requirements=window_requirements,
        initial_condition=PrognosticState(initial_condition),
    )
    assert "foo" in data.variable_metadata.keys()
    assert "foo2" in data.variable_metadata.keys()


@pytest.mark.parametrize(
    "timestamps, expected_indices",
    [
        (
            [
                "2020-01-01T00:00:00",
                "2020-01-02T00:00:00",
            ],
            [0, 2],
        ),
        (
            [
                "2020-01-01T00:00:00",
                "2021-01-02T00:00:00",
            ],
            None,
        ),
        (
            [
                "2021-01-02T00:00:00",
            ],
            None,
        ),
    ],
)
def test_TimestampList_as_indices(timestamps, expected_indices):
    time_index = xr.CFTimeIndex(
        [
            cftime.DatetimeJulian(2020, 1, 1, 0, 0, 0),
            cftime.DatetimeJulian(2020, 1, 1, 12, 0, 0),
            cftime.DatetimeJulian(2020, 1, 2, 0, 0, 0),
            cftime.DatetimeJulian(2020, 1, 2, 12, 0, 0),
        ]
    )
    timestamp_list = TimestampList(timestamps)
    if expected_indices is None:
        with pytest.raises(ValueError):
            timestamp_list.as_indices(time_index)
    else:
        np.testing.assert_equal(
            timestamp_list.as_indices(time_index), np.array(expected_indices)
        )


def test_inference_data_with_perturbations(tmp_path):
    _create_dataset_on_disk(tmp_path, n_times=14)
    batch_size = 1
    step = 7
    config = InferenceDataLoaderConfig(
        dataset=XarrayDataConfig(
            data_path=tmp_path,
            n_repeats=1,
        ),
        start_indices=InferenceInitialConditionIndices(
            first=0, n_initial_conditions=batch_size, interval=step
        ),
        perturbations=SSTPerturbation(
            sst=[PerturbationSelector(type="constant", config={"amplitude": 2.0})]
        ),
    )
    n_forward_steps_in_memory = 3
    original_foo = xr.open_dataset(
        os.path.join(tmp_path, "data.nc"), decode_times=False
    )["foo"].values[0 : n_forward_steps_in_memory + 1, :, :]
    window_requirements = DataRequirements(
        names=["foo", "constant_mask"],
        n_timesteps=n_forward_steps_in_memory + 1,
    )
    initial_condition_requirements = PrognosticStateDataRequirements(
        names=["foo"],
        n_timesteps=1,
    )
    data = get_inference_data(
        config,
        total_forward_steps=6,
        window_requirements=window_requirements,
        initial_condition=initial_condition_requirements,
        surface_temperature_name="foo",
        ocean_fraction_name="constant_mask",
    )
    batch_data = next(iter(data.loader))
    np.testing.assert_allclose(
        original_foo + 2.0,
        batch_data.data["foo"].cpu().numpy()[0, :, :, :],
    )
    np.testing.assert_allclose(
        original_foo[:1, :, :] + 2.0,
        data.initial_condition.as_batch_data().data["foo"].cpu().numpy()[0, :, :, :],
    )


def test_inference_persistence_names(tmp_path):
    _create_dataset_on_disk(tmp_path, n_times=14)

    config = InferenceDataLoaderConfig(
        dataset=XarrayDataConfig(data_path=tmp_path),
        start_indices=ExplicitIndices([0, 3]),
        persistence_names=["foo"],
    )
    window_requirements = DataRequirements(
        names=["foo", "bar"],
        n_timesteps=3,
    )
    dataset = InferenceDataset(
        config,
        9,
        requirements=window_requirements,
    )
    first_item = dataset[0].data
    second_item = dataset[1].data
    # ensure first and second time steps are the same
    torch.testing.assert_close(first_item["foo"][:, 0], first_item["foo"][:, 1])
    # ensure the entire first and second returned items
    torch.testing.assert_close(first_item["foo"], second_item["foo"])
    # ensure this is not the case for another variable
    assert not torch.all(first_item["bar"] == second_item["bar"])


def test_zarr_engine_used_sequence():
    config = DataLoaderConfig(
        dataset=ConcatDatasetConfig(
            concat=[
                XarrayDataConfig(data_path="foo", file_pattern=".zarr", engine="zarr"),
                XarrayDataConfig(data_path="bar"),
            ]
        ),
        batch_size=1,
    )
    assert config.zarr_engine_used


def test_zarr_engine_used_mapping():
    config = DataLoaderConfig(
        dataset=MergeDatasetConfig(
            merge=[
                ConcatDatasetConfig(
                    concat=[
                        XarrayDataConfig(
                            data_path="foo", file_pattern=".zarr", engine="zarr"
                        ),
                        XarrayDataConfig(data_path="bar"),
                    ]
                ),
                XarrayDataConfig(data_path="bar"),
            ]
        ),
        batch_size=1,
    )
    assert config.zarr_engine_used


def test_data_loader_maintains_backward_concat_compatibility(recwarn):
    data_loader = DataLoaderConfig(
        dataset=[XarrayDataConfig(data_path="some_path")],
        batch_size=1,
    )
    assert isinstance(data_loader.dataset, ConcatDatasetConfig)
    warning = recwarn.pop(DeprecationWarning)
    assert "Dataset list format is deprecated" in str(warning.message)


@pytest.mark.parametrize(
    (
        "shuffle",
        "steps_on_disk",
        "n_timesteps_requirement",
        "time_buffer",
        "expected_start_indices",
    ),
    (
        pytest.param(
            False,
            7,
            2,
            0,
            list(range(0, 6)),
            id="time_buffer=0",
        ),
        pytest.param(
            False,
            7,
            2,
            2,
            list(range(0, 6)),
            id="time_buffer=2",
        ),
        pytest.param(
            False,
            10,
            2,
            2,
            list(range(0, 9)),
            id="time_buffer=2, steps_on_disk=10",
        ),
        pytest.param(
            False,
            6,
            2,
            2,
            list(range(0, 3)),
            id=(
                "time_buffer=2, steps_on_disk=6 "
                "(lose some samples due to dropping last 'preloaded' batch)"
            ),
        ),
        pytest.param(
            False,
            8,
            3,
            2,
            list(range(0, 6)),
            id="time_buffer=2, steps_on_disk=10, n_timesteps_requirement=3",
        ),
        pytest.param(
            False,
            10,
            3,
            2,
            list(range(0, 6)),
            id=(
                "time_buffer=2, steps_on_disk=10, n_timesteps_requirement=3 "
                "(lose some samples due to dropping last 'preloaded' batch)"
            ),
        ),
        pytest.param(
            True,
            13,
            2,
            2,
            list(range(0, 12)),
            id="time_buffer=2, shuffle=True",
        ),
    ),
)
def test_time_buffer(
    tmp_path,
    shuffle,
    steps_on_disk,
    n_timesteps_requirement,
    time_buffer,
    expected_start_indices,
):
    _create_dataset_on_disk(tmp_path, n_times=steps_on_disk)
    dataset_times = xr.open_dataset(
        os.path.join(tmp_path, "data.nc"),
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
    )["time"].values
    window_size = time_buffer + 1
    i_window_starts = [
        i
        for i in range(0, len(dataset_times), window_size)
        if i + window_size <= len(dataset_times) - n_timesteps_requirement + 1
    ]
    dataset_window_times = [dataset_times[i : i + window_size] for i in i_window_starts]
    config = DataLoaderConfig(
        dataset=XarrayDataConfig(data_path=tmp_path),
        batch_size=1,
        num_data_workers=0,
        time_buffer=time_buffer,
    )
    requirements = DataRequirements(["foo"], n_timesteps_requirement)
    data = get_gridded_data(config, shuffle, requirements)
    assert len(data.loader) == len(expected_start_indices)
    start_times = []
    for batch in data.loader:
        start_times.append(batch.time[0, 0].item())
        assert batch.n_timesteps == n_timesteps_requirement
    if shuffle:
        assert set(start_times) == set(dataset_times[expected_start_indices])
    else:
        assert start_times == list(dataset_times[expected_start_indices])
    for i_window_start in i_window_starts:
        # confirm that a "window's worth" of data is loaded consecutively;
        # with shuffling we know that it will match one window, but not which one
        loaded_times = start_times[i_window_start : i_window_start + window_size]
        window_matches = [
            set(loaded_times) == set(window_times)
            for window_times in dataset_window_times
        ]
        assert sum(window_matches) == 1



import dataclasses
import re
from collections.abc import Callable

from torch import nn

from fme.core.typing_ import TensorDict, TensorMapping

LEVEL_PATTERN = re.compile(r"_(\d+)$")
TEMPLATE = "{name}{suffix}"


StepMethod = Callable[
    [TensorMapping, TensorMapping, Callable[[nn.Module], nn.Module]], TensorDict
]


def get_multi_call_name(name: str, suffix: str) -> str:
    """Get multi-call name, appropriately handling 3D variables.

    For 2D fields this trivially appends the suffix to the name; for 3D fields
    we insert the suffix between the variable name and vertical level label,
    following how these variables are pre-processed.

    Args:
       name: Name of field.
       suffix: Suffix to append indicating multi-call multiplier.

    Returns:
       Name of multi-call field associated with the given suffix.

    Examples:
       >>> fme.core.multi_call.get_multi_call_name("foo", "_with_quartered_co2")
       'foo_with_quartered_co2'
       >>> fme.core.multi_call.get_multi_call_name("bar_0", "_with_quartered_co2")
       'bar_with_quartered_co2_0'
    """
    match = LEVEL_PATTERN.search(name)
    if match is not None:
        name = name[: match.start()]
        suffix = suffix + match.group(0)
    return TEMPLATE.format(name=name, suffix=suffix)


@dataclasses.dataclass
class MultiCallConfig:
    """Configuration for doing 'multi-call' predictions where an input variable (e.g.
    CO2) is varied by multiplying by floats and then certain output variables (e.g.
    radiative heating or fluxes) are predicted.

    Parameters:
        forcing_name: name of the variable to perturb in the forcing data, e.g. "co2".
        forcing_multipliers: mapping from a label suffix to a multiplier that is applied
            to the 'forcing_name' variable. For example, could be
            {"_quadrupled_co2": 4, "_halved_co2": 0.5}. The suffixes will be appended to
            the output_names below.
        output_names: names of the variables to predict given perturbed forcing. For
            example, ["ULWRFtoa", "USWRFsfc"].
    """

    forcing_name: str
    forcing_multipliers: dict[str, float]
    output_names: list[str]

    def __post_init__(self):
        self._names = []
        for name in self.output_names:
            self._names.extend(self.get_multi_called_names(name))

    def get_multi_called_names(self, name: str) -> list[str]:
        names = []
        for suffix in self.forcing_multipliers:
            names.append(get_multi_call_name(name, suffix))
        return names

    def validate(self, in_names: list[str], out_names: list[str]):
        if self.forcing_name not in in_names:
            raise ValueError(
                f"forcing name {self.forcing_name} not in input names. It is required "
                "as a forcing given provided radiation multi call configuration."
            )
        if self.forcing_name in out_names:
            raise ValueError(
                f"forcing name {self.forcing_name} is in the output names, "
                "but it must be a forcing variable, not an output."
            )
        for name in self.output_names:
            if name not in out_names:
                raise ValueError(
                    f"{name} not in output names. It is required "
                    "as an output given provided radiation multi call configuration."
                )
        for multi_called_name in self.names:
            if multi_called_name in in_names:
                raise ValueError(
                    f"The multi-call output {multi_called_name} is already in in_names."
                    " This will lead to a conflict--please rename the input or "
                    "use a different multi-call suffix label."
                )
            if multi_called_name in out_names:
                raise ValueError(
                    f"The multi-call output {multi_called_name} is already in "
                    "out_names. This will lead to a conflict--please rename the output "
                    "or use a different multi-call suffix label."
                )

    @property
    def names(self) -> list[str]:
        """
        Return the names of all multi-called output variables,
        often radiative fluxes.

        E.g. ['ULWRFtoa_quadrupled_co2'].
        """
        return self._names

    def build(self, step_method: StepMethod) -> "MultiCall":
        return MultiCall(self, step_method)


class MultiCall:
    """Class for doing 'multi-call' predictions where a forcing variable is varied
    and certain outputs are saved for each value.

    Given a 'step' method that takes the input data and next-step forcing data, this
    class will call the step method multiple times, changing the value of the desired
    forcing variable each time. Specified outputs are saved for each value with their
    names modified to indicate the forcing value cahnge.
    """

    def __init__(
        self,
        config: MultiCallConfig,
        step_method: StepMethod,
    ):
        self.forcing_name = config.forcing_name
        self.forcing_multipliers = config.forcing_multipliers
        self.output_names = config.output_names
        self._names = config.names
        self._step = step_method

    @property
    def names(self) -> list[str]:
        return self._names

    def step(
        self,
        input: TensorMapping,
        next_step_forcing_data: TensorMapping,
        wrapper: Callable[[nn.Module], nn.Module] = lambda x: x,
    ) -> TensorDict:
        predictions = {}
        unscaled_forcing = input[self.forcing_name]
        for suffix, multiplier in self.forcing_multipliers.items():
            scaled_input = dict(input)
            scaled_input[self.forcing_name] = multiplier * unscaled_forcing
            output = self._step(scaled_input, next_step_forcing_data, wrapper)

            for name in self.output_names:
                new_name = get_multi_call_name(name, suffix)
                predictions[new_name] = output[name]

        return predictions



# flake8: noqa
# Copied from https://github.com/ai2cm/modulus/commit/22df4a9427f5f12ff6ac891083220e7f2f54d229
# Copyright (c) 2023, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from torch import nn


class ComplexReLU(nn.Module):
    """
    Complex-valued variants of the ReLU activation function
    """

    def __init__(self, negative_slope=0.0, mode="real", bias_shape=None, scale=1.0):
        super(ComplexReLU, self).__init__()

        # store parameters
        self.mode = mode
        if self.mode in ["modulus", "halfplane"]:
            if bias_shape is not None:
                self.bias = nn.Parameter(
                    scale * torch.ones(bias_shape, dtype=torch.float32)
                )
            else:
                self.bias = nn.Parameter(scale * torch.ones((1), dtype=torch.float32))
        else:
            self.bias = 0

        self.negative_slope = negative_slope
        self.act = nn.LeakyReLU(negative_slope=negative_slope)

    def forward(self, z: torch.Tensor) -> torch.Tensor:
        if self.mode == "cartesian":
            zr = torch.view_as_real(z)
            za = self.act(zr)
            out = torch.view_as_complex(za)

        elif self.mode == "modulus":
            zabs = torch.sqrt(torch.square(z.real) + torch.square(z.imag))
            out = torch.where(zabs + self.bias > 0, (zabs + self.bias) * z / zabs, 0.0)
            # out = self.act(zabs - self.bias) * torch.exp(1.j * z.angle())

        elif self.mode == "halfplane":
            # bias is an angle parameter in this case
            modified_angle = torch.angle(z) - self.bias
            condition = torch.logical_and(
                (0.0 <= modified_angle), (modified_angle < torch.pi / 2.0)
            )
            out = torch.where(condition, z, self.negative_slope * z)

        elif self.mode == "real":
            zr = torch.view_as_real(z)
            outr = zr.clone()
            outr[..., 0] = self.act(zr[..., 0])
            out = torch.view_as_complex(outr)

        else:
            raise NotImplementedError

        return out


class ComplexActivation(nn.Module):
    """
    A module implementing complex-valued activation functions.
    The module supports different modes of operation, depending on how
    the complex numbers are treated for the activation function:
    - "cartesian": the activation function is applied separately to the
       real and imaginary parts of the complex input.
    - "modulus": the activation function is applied to the modulus of
       the complex input, after adding a learnable bias.
    - any other mode: the complex input is returned as-is (identity operation).
    """

    def __init__(self, activation, mode="cartesian", bias_shape=None):
        super(ComplexActivation, self).__init__()

        # store parameters
        self.mode = mode
        if self.mode == "modulus":
            if bias_shape is not None:
                self.bias = nn.Parameter(torch.zeros(bias_shape, dtype=torch.float32))
            else:
                self.bias = nn.Parameter(torch.zeros((1), dtype=torch.float32))
        else:
            bias = torch.zeros((1), dtype=torch.float32)
            self.register_buffer("bias", bias)

        # real valued activation
        self.act = activation

    def forward(self, z: torch.Tensor) -> torch.Tensor:
        if self.mode == "cartesian":
            zr = torch.view_as_real(z)
            za = self.act(zr)
            out = torch.view_as_complex(za)
        elif self.mode == "modulus":
            zabs = torch.sqrt(torch.square(z.real) + torch.square(z.imag))
            out = self.act(zabs + self.bias) * torch.exp(1.0j * z.angle())
        else:
            # identity
            out = z

        return out
